import unittest
import logging
import os
import numpy as np
from medcat.cdb_maker import CDBMaker
from medcat.cdb import CDB
from medcat.config import Config
from medcat.preprocessing.cleaners import prepare_name

#cdb.csv
#cui  name  ontologies  name_status type_ids  description
#C0000039 Virus MSH P T109|T123 Synthetic phospholipid used in liposomes and lipid bilayers to study biological membranes. It is also a major constituent of PULMONARY SURFACTANTS.
#C0000039 Virus M     T234  
#C0000039 Virus M |Virus K|Virus Z        
#C0000139 Virus M|Virus K|Virus Z   P   
#C0000139 Virus   A   

#cdb2.csv
#cui  name  ontologies  name_status type_ids  description
#C0000239 Second csv        

#TESTS RUN IN ALPHABETICAL ORDER - CONTROLLING WITH '[class_letter]Class and test_[classletter subclassletter]' function syntax


class A_CDBMakerLoadTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print("Load test database csvs for load tests")
        config = Config()
        config.general['log_level'] = logging.DEBUG
        config.general["spacy_model"] = "en_core_web_md"
        cls.maker = CDBMaker(config)
        csvs = [
            os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'examples', 'cdb.csv'),
            os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'examples', 'cdb_2.csv')
        ]
        cls.cdb = cls.maker.prepare_csvs(csvs, full_build=True)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.maker.destroy_pipe()

    def test_aa_cdb_names_length(self):
        self.assertEqual(len(self.cdb.cui2names), 3, "Should equal 3")

    def test_ab_cdb_names_output(self):
        target_result = {'C0000039': {'virus~k', 'virus', 'virus~m', 'virus~z'}, 'C0000139': {'virus~k', 'virus', 'virus~m', 'virus~z'}, 'C0000239': {'second~csv'}}
        self.assertEqual(self.cdb.cui2names, target_result)

    def test_ac_cdb_snames_length(self):
        self.assertEqual(len(self.cdb.cui2snames), 3, "Should equal 3")

    def test_ad_cdb_snames_output(self):
        target_result = {'C0000039': {'virus~k', 'virus', 'virus~m', 'virus~z'}, 'C0000139': {'virus~k', 'virus', 'virus~m', 'virus~z'}, 'C0000239': {'second', 'second~csv'}}
        self.assertEqual(self.cdb.cui2snames, target_result)

    def test_ae_cdb_name_to_cuis_length(self):
        self.assertEqual(len(self.cdb.name2cuis), 5, "Should equal 5")

    def test_af_cdb_name_to_cuis_output(self):
        target_result = {'virus': ['C0000039', 'C0000139'], 'virus~m': ['C0000039', 'C0000139'], 'virus~k': ['C0000039', 'C0000139'], 'virus~z': ['C0000039', 'C0000139'], 'second~csv': ['C0000239']}
        self.assertEqual(self.cdb.name2cuis, target_result)

    def test_ag_cdb_cuis_to_tags_length(self):
        self.assertEqual(len(self.cdb.cui2tags), 0, "Should equal 0")

    def test_ah_cdb_cuis_to_tags_output(self):
        target_result = {}
        self.assertEqual(self.cdb.cui2tags, target_result)

    def test_ai_cdb_cui_to_preferred_name_length(self):
        self.assertEqual(len(self.cdb.cui2preferred_name), 2, "Should equal 2")

    def test_aj_cdb_cui_to_preferred_name_output(self):
        target_result = {'C0000039': 'Virus', 'C0000139': 'Virus Z'}
        self.assertEqual(self.cdb.cui2preferred_name, target_result)

    def test_ak_cdb_cui_to_context_vectors_length(self):
        self.assertEqual(len(self.cdb.cui2context_vectors), 0, "Should equal 0")

    def test_al_cdb_cui_to_context_vectors_output(self):
        target_result = {}
        self.assertEqual(self.cdb.cui2context_vectors, target_result)

    def test_am_cdb_cui_to_count_train_length(self):
        self.assertEqual(len(self.cdb.cui2count_train), 0, "Should equal 0")

    def test_an_cdb_cui_to_count_train_output(self):
        target_result = {}
        self.assertEqual(self.cdb.cui2count_train, target_result)

    def test_ao_cdb_name_to_cui_to_status_length(self):
        self.assertEqual(len(self.cdb.name2cuis2status), 5, "Should equal 5")

    def test_ap_cdb_name_to_cui_to_status_output(self):
        target_result = {'virus': {'C0000039': 'P', 'C0000139': 'A'}, 'virus~m': {'C0000039': 'A', 'C0000139': 'P'}, 'virus~k': {'C0000039': 'A', 'C0000139': 'P'}, 'virus~z': {'C0000039': 'A', 'C0000139': 'P'}, 'second~csv': {'C0000239': 'A'}}
        self.assertEqual(self.cdb.name2cuis2status, target_result)

    def test_aq_cdb_cui_to_type_ids_length(self):
        self.assertEqual(len(self.cdb.cui2type_ids), 3, "Should equal 3")

    def test_ar_cdb_cui_to_type_ids_output(self):
        target_result = {'C0000039': {'T234', 'T109', 'T123'}, 'C0000139': set(), 'C0000239': set()}
        self.assertEqual(self.cdb.cui2type_ids, target_result)

    def test_as_cdb_additional_info_length(self):
        self.assertEqual(len(self.cdb.addl_info), 8, "Should equal 8")

    def test_at_cdb_additional_info_output(self):
        target_result = {'cui2icd10': {}, 'cui2opcs4': {}, 'cui2ontologies': {'C0000039': {'MSH'}}, 'cui2original_names': {'C0000039': {'Virus K', 'Virus M', 'Virus', 'Virus Z'}, 'C0000139': {'Virus K', 'Virus M', 'Virus', 'Virus Z'}, 'C0000239': {'Second csv'}}, 'cui2description': {'C0000039': 'Synthetic phospholipid used in liposomes and lipid bilayers to study biological membranes. It is also a major constituent of PULMONARY SURFACTANTS.'}, 'type_id2name': {}, 'type_id2cuis': {'T109': {'C0000039'}, 'T123': {'C0000039'}, 'T234': {'C0000039'}}, 'cui2group': {}}
        self.assertEqual(self.cdb.addl_info, target_result)


class B_CDBMakerEditTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print("Load test database csvs for edit tests")
        cls.config = Config()
        cls.config.general['log_level'] = logging.DEBUG
        cls.config.general["spacy_model"] = "en_core_web_md"
        cls.maker = CDBMaker(cls.config)
        csvs = [
            os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'examples', 'cdb.csv'),
            os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'examples', 'cdb_2.csv')
        ]
        cls.cdb = cls.maker.prepare_csvs(csvs, full_build=True)
        cls.cdb2 = CDB(cls.config)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.maker.destroy_pipe()

    def test_ba_addition_of_new_name(self):
        self.cdb.add_names(cui='C0000239', names=prepare_name('MY: new,-_! Name.', self.maker.pipe.spacy_nlp, {}, self.config), name_status='P', full_build=True)
        self.assertEqual(len(self.cdb.name2cuis), 6, "Should equal 6")
        target_result = {'MY: new,-_! Name.', 'Second csv'}
        self.assertEqual(self.cdb.addl_info['cui2original_names']['C0000239'], target_result)
        self.assertIn('my~:~new~name~.', self.cdb.name2cuis)
        self.assertIn('my~:~new', self.cdb.snames)
        self.assertIn('my~:~new~name~.', self.cdb.name2cuis2status)

    def test_bb_removal_of_name(self):
        self.cdb.remove_names(cui='C0000239', names=prepare_name('MY: new,-_! Name.', self.maker.pipe.spacy_nlp, {}, self.config))
        self.assertEqual(len(self.cdb.name2cuis), 5, "Should equal 5")
        self.assertNotIn('my:newname.', self.cdb.name2cuis2status)

    def test_bc_filter_by_cui(self):
        cuis_to_keep = {'C0000039'}
        self.cdb.filter_by_cui(cuis_to_keep=cuis_to_keep)
        self.assertEqual(len(self.cdb.cui2names), 2, "Should equal 2")
        self.assertEqual(len(self.cdb.name2cuis), 4, "Should equal 4")
        self.assertEqual(len(self.cdb.snames), 4, "Should equal 4")

    def test_bd_addition_of_context_vector_positive(self):
        np.random.seed(11)
        cuis = list(self.cdb.cui2names.keys())
        for i in range(2):
                for cui in cuis:
                        vectors = {}
                        for cntx_type in self.config.linking['context_vector_sizes']:
                                vectors[cntx_type] = np.random.rand(300)
                        self.cdb.update_context_vector(cui, vectors, negative=False)

        self.assertEqual(self.cdb.cui2count_train['C0000139'], 2, "Count should equal 2")
        self.assertEqual(self.cdb.cui2context_vectors['C0000139']['long'].shape[0], 300, "Dimensions should equal 300")

    def test_be_addition_of_context_vector_negative(self):
        np.random.seed(11)
        cuis = list(self.cdb.cui2names.keys())
        for i in range(2):
                for cui in cuis:
                        vectors = {}
                        for cntx_type in self.config.linking['context_vector_sizes']:
                                vectors[cntx_type] = np.random.rand(300)
                        self.cdb.update_context_vector(cui, vectors, negative=True)

        self.assertEqual(self.cdb.cui2count_train['C0000139'], 2, "Count should equal 2")
        self.assertEqual(self.cdb.cui2context_vectors['C0000139']['long'].shape[0], 300, "Dimensions should equal 300")

    def test_bf_import_training(self):
        self.cdb2.import_training(cdb=self.cdb, overwrite=True)
        self.assertEqual(self.cdb.cui2count_train['C0000139'], 2, "Count should equal 2")
        self.assertEqual(self.cdb.cui2context_vectors['C0000139']['long'].shape[0], 300, "Dimensions should equal 300")


    def test_bg_save_and_load_model_context_vectors(self):
        self.cdb.save("./tmp_cdb.dat")
        self.cdb2 = CDB.load('./tmp_cdb.dat')
        self.assertEqual(self.cdb.cui2count_train['C0000139'], 2, "Count should equal 2")
        self.assertEqual(self.cdb.cui2context_vectors['C0000139']['long'].shape[0], 300, "Dimensions should equal 300")


    def test_bh_reset_training(self):
        self.cdb.reset_training()
        target_result = {}
        self.assertEqual(self.cdb.cui2context_vectors, target_result)


if __name__ == '__main__':
    unittest.main()



import logging
import unittest
import numpy as np
from medcat.cdb import CDB
from medcat.cdb_maker import CDBMaker
from medcat.config import Config
from medcat.preprocessing.cleaners import prepare_name


class CdbMakerArchiveTests(unittest.TestCase):

    def setUp(self):
        self.config = Config()
        self.config.general['log_level'] = logging.DEBUG
        self.maker = CDBMaker(self.config)

        # Building a new CDB from two files (full_build)
        csvs = ['../examples/cdb.csv', '../examples/cdb_2.csv']
        self.cdb = self.maker.prepare_csvs(csvs, full_build=True)

    def test_prepare_csvs(self):
        assert len(self.cdb.cui2names) == 3
        assert len(self.cdb.cui2snames) == 3
        assert len(self.cdb.name2cuis) == 5
        assert len(self.cdb.cui2tags) == 3
        assert len(self.cdb.cui2preferred_name) == 2
        assert len(self.cdb.cui2context_vectors) == 3
        assert len(self.cdb.cui2count_train) == 3
        assert self.cdb.name2cuis2status['virus']['C0000039'] == 'P'
        assert self.cdb.cui2type_ids['C0000039'] == {'T234', 'T109', 'T123'}
        assert self.cdb.addl_info['cui2original_names']['C0000039'] == {'Virus', 'Virus K', 'Virus M', 'Virus Z'}
        assert self.cdb.addl_info['cui2description']['C0000039'].startswith("Synthetic")

    def test_name_addition(self):
        self.cdb.add_names(cui='C0000239', names=prepare_name('MY: new,-_! Name.', self.maker.pipe.get_spacy_nlp(), {}, self.config), name_status='P', full_build=True)
        assert self.cdb.addl_info['cui2original_names']['C0000239'] == {'MY: new,-_! Name.', 'Second csv'}
        assert 'my:newname.' in self.cdb.name2cuis
        assert 'my:new' in self.cdb.snames
        assert 'my:newname.' in self.cdb.name2cuis2status
        assert self.cdb.name2cuis2status['my:newname.'] == {'C0000239': 'P'}

    def test_name_removal(self):
        self.cdb.remove_names(cui='C0000239', names=prepare_name('MY: new,-_! Name.', self.maker.pipe.get_spacy_nlp(), {}, self.config))
        # Run again to make sure it does not break anything
        self.cdb.remove_names(cui='C0000239', names=prepare_name('MY: new,-_! Name.', self.maker.pipe.get_spacy_nlp(), {}, self.config))
        assert len(self.cdb.name2cuis) == 5
        assert 'my:newname.' not in self.cdb.name2cuis2status

    def test_filtering(self):
        cuis_to_keep = {'C0000039'} # Because of transition 2 will be kept
        self.cdb.filter_by_cui(cuis_to_keep=cuis_to_keep)
        assert len(self.cdb.cui2names) == 2
        assert len(self.cdb.name2cuis) == 4
        assert len(self.cdb.snames) == 4

    def test_vector_addition(self):
        self.cdb.reset_training()
        np.random.seed(11)
        cuis = list(self.cdb.cui2names.keys())
        for i in range(2):
            for cui in cuis:
                vectors = {}
                for cntx_type in self.config.linking['context_vector_sizes']:
                    vectors[cntx_type] = np.random.rand(300)
                self.cdb.update_context_vector(cui, vectors, negative=False)

        assert self.cdb.cui2count_train['C0000139'] == 2
        assert self.cdb.cui2context_vectors['C0000139']['long'].shape[0] == 300


    def test_negative(self):
        cuis = list(self.cdb.cui2names.keys())
        for cui in cuis:
            vectors = {}
            for cntx_type in self.config.linking['context_vector_sizes']:
                vectors[cntx_type] = np.random.rand(300)
            self.cdb.update_context_vector(cui, vectors, negative=True)

        assert self.cdb.cui2count_train['C0000139'] == 2
        assert self.cdb.cui2context_vectors['C0000139']['long'].shape[0] == 300

    def test_save_and_load(self):
        self.cdb.save("./tmp_cdb.dat")
        cdb2 = CDB.load('./tmp_cdb.dat')
        # Check a random thing
        assert cdb2.cui2context_vectors['C0000139']['long'][7] == self.cdb.cui2context_vectors['C0000139']['long'][7]

    def test_training_import(self):
        cdb2 = CDB.load('./tmp_cdb.dat')
        self.cdb.reset_training()
        cdb2.reset_training()
        np.random.seed(11)
        cuis = list(self.cdb.cui2names.keys())
        for i in range(2):
            for cui in cuis:
                vectors = {}
                for cntx_type in self.config.linking['context_vector_sizes']:
                    vectors[cntx_type] = np.random.rand(300)
                self.cdb.update_context_vector(cui, vectors, negative=False)

        cdb2.import_training(cdb=self.cdb, overwrite=True)
        assert cdb2.cui2context_vectors['C0000139']['long'][7] == self.cdb.cui2context_vectors['C0000139']['long'][7]
        assert cdb2.cui2count_train['C0000139'] == self.cdb.cui2count_train['C0000139']

    def test_concept_similarity(self):
        cdb = CDB(config=self.config)
        np.random.seed(11)
        for i in range(500):
            cui = "C" + str(i)
            type_ids = {'T-' + str(i%10)}
            cdb.add_concept(cui=cui, names=prepare_name('Name: ' + str(i), self.maker.pipe.get_spacy_nlp(), {}, self.config), ontologies=set(),
                            name_status='P', type_ids=type_ids, description='', full_build=True)

            vectors = {}
            for cntx_type in self.config.linking['context_vector_sizes']:
                vectors[cntx_type] = np.random.rand(300)
            cdb.update_context_vector(cui, vectors, negative=False)
        res = cdb.most_similar('C200', 'long', type_id_filter=['T-0'], min_cnt=1, topn=10, force_build=True)
        assert len(res) == 10

    def test_training_reset(self):
        self.cdb.reset_training()
        assert len(self.cdb.cui2context_vectors['C0']) == 0
        assert self.cdb.cui2count_train['C0'] == 0



import os
import shutil
import pickle
import traceback
import json
import logging
import math
import time
import psutil
from time import sleep
from copy import deepcopy
from multiprocess import Process, Manager, cpu_count
from multiprocess.queues import Queue
from multiprocess.synchronize import Lock
from typing import Union, List, Tuple, Optional, Dict, Iterable, Set
from itertools import islice, chain, repeat
from datetime import date
from tqdm.autonotebook import tqdm, trange
from spacy.tokens import Span, Doc, Token
from spacy.language import Language

from medcat import __version__
from medcat.preprocessing.tokenizers import spacy_split_all
from medcat.pipe import Pipe
from medcat.preprocessing.taggers import tag_skip_and_punct
from medcat.cdb import CDB
from medcat.utils.matutils import intersect_nonempty_set
from medcat.utils.loggers import add_handlers
from medcat.utils.data_utils import make_mc_train_test, get_false_positives
from medcat.utils.normalizers import BasicSpellChecker
from medcat.utils.checkpoint import Checkpoint, CheckpointConfig, CheckpointManager
from medcat.utils.helpers import tkns_from_doc, get_important_config_parameters
from medcat.utils.hasher import Hasher
from medcat.ner.vocab_based_ner import NER
from medcat.linking.context_based_linker import Linker
from medcat.utils.filters import get_project_filters, check_filters
from medcat.preprocessing.cleaners import prepare_name
from medcat.meta_cat import MetaCAT
from medcat.utils.meta_cat.data_utils import json_to_fake_spacy
from medcat.config import Config
from medcat.vocab import Vocab
from medcat.utils.decorators import deprecated


class CAT(object):
    r'''
    The main MedCAT class used to annotate documents, it is built on top of spaCy
    and works as a spaCy pipline. Creates an instance of a spaCy pipline that can
    be used as a spacy nlp model.

    Args:
        cdb (medcat.cdb.CDB):
            The concept database that will be used for NER+L
        config (medcat.config.Config):
            Global configuration for medcat
        vocab (medcat.vocab.Vocab, optional):
            Vocabulary used for vector embeddings and spelling. Default: None
        meta_cats (list of medcat.meta_cat.MetaCAT, optional):
            A list of models that will be applied sequentially on each
            detected annotation.

    Attributes (limited):
        cdb (medcat.cdb.CDB):
            Concept database used with this CAT instance, please do not assign
            this value directly.
        config (medcat.config.Config):
            The global configuration for medcat. Usually cdb.config will be used for this
            field.
        vocab (medcat.utils.vocab.Vocab):
            The vocabulary object used with this instance, please do not assign
            this value directly.
        config - WILL BE REMOVED - TEMPORARY PLACEHOLDER

    Examples:
        >>>cat = CAT(cdb, vocab)
        >>>spacy_doc = cat("Put some text here")
        >>>print(spacy_doc.ents) # Detected entites
    '''
    log = logging.getLogger(__package__)
    # Add file and console handlers
    log = add_handlers(log)
    DEFAULT_MODEL_PACK_NAME = "medcat_model_pack"

    def __init__(self,
                 cdb: CDB,
                 vocab: Vocab,
                 config: Optional[Config] = None,
                 meta_cats: List[MetaCAT] = []) -> None:
        self.cdb = cdb
        self.vocab = vocab
        if config is None:
            # Take config from the cdb
            self.config = cdb.config
        else:
            # Take the new config and assign it to the CDB also

            self.config = config
            self.cdb.config = config
        self._meta_cats = meta_cats
        self._create_pipeline(self.config)

    def _create_pipeline(self, config):
        # Set log level
        self.log.setLevel(config.general['log_level'])

        # Build the pipeline
        self.pipe = Pipe(tokenizer=spacy_split_all, config=config)
        self.pipe.add_tagger(tagger=tag_skip_and_punct,
                             name='skip_and_punct',
                             additional_fields=['is_punct'])

        spell_checker = BasicSpellChecker(cdb_vocab=self.cdb.vocab, config=config, data_vocab=self.vocab)
        self.pipe.add_token_normalizer(spell_checker=spell_checker, config=config)

        # Add NER
        self.ner = NER(self.cdb, config)
        self.pipe.add_ner(self.ner)

        # Add LINKER
        self.linker = Linker(self.cdb, self.vocab, config)
        self.pipe.add_linker(self.linker)

        # Add meta_annotaiton classes if they exist
        for meta_cat in self._meta_cats:
            self.pipe.add_meta_cat(meta_cat, meta_cat.config.general['category_name'])

        # Set max document length
        self.pipe.spacy_nlp.max_length = config.preprocessing.get('max_document_length', 1000000)

    @deprecated(message="Replaced with cat.pipe.spacy_nlp.")
    def get_spacy_nlp(self) -> Language:
        ''' Returns the spacy pipeline with MedCAT
        '''
        return self.pipe.spacy_nlp

    def get_hash(self):
        r''' Will not be a deep hash but will try to cactch all the changing parts during training.
        '''
        hasher = Hasher()
        hasher.update(self.cdb.get_hash())

        hasher.update(self.config.get_hash())

        for mc in self._meta_cats:
            hasher.update(mc.get_hash())

        return hasher.hexdigest()

    def get_model_card(self, as_dict=False):
        card = {
                'Model ID': self.config.version['id'],
                'Last Modifed On': self.config.version['last_modified'],
                'History (from least to most recent)': self.config.version['history'],
                'Description': self.config.version['description'],
                'Source Ontology': self.config.version['ontology'],
                'Location': self.config.version['location'],
                'MetaCAT models': self.config.version['meta_cats'],
                'Basic CDB Stats': self.config.version['cdb_info'],
                'Performance': self.config.version['performance'],
                'Important Parameters (Partial view, all available in cat.config)': get_important_config_parameters(self.config),
                'MedCAT Version': self.config.version['medcat_version']
                }

        if as_dict:
            return card
        else:
            return json.dumps(card, indent=2, sort_keys=False)

    def _versioning(self):
        # Check version info and do not allow without it
        if self.config.version['description'] == 'No description':
            self.log.warning("Please consider populating the version information [description, performance, location, ontology] in cat.config.version")

        # Fill the stuff automatically that is needed for versioning
        m = self.get_hash()
        version = self.config.version
        if version['id'] is None or m != version['id']:
            if version['id'] is not None:
                version['history'].append(version['id'])
            version['id'] = m
            version['last_modified'] = date.today().strftime("%d %B %Y")
            version['cdb_info'] = self.cdb._make_stats()
            version['meta_cats'] = {meta_cat.config.general['category_name']: meta_cat.config.general['description'] for meta_cat in self._meta_cats}
            version['medcat_version'] = __version__
            self.log.warning("Please consider updating [description, performance, location, ontology] in cat.config.version")

    def create_model_pack(self, save_dir_path: str, model_pack_name: str = DEFAULT_MODEL_PACK_NAME) -> str:
        r''' Will crete a .zip file containing all the models in the current running instance
        of MedCAT. This is not the most efficient way, for sure, but good enough for now.

        model_pack_name - an id will be appended to this name

        returns:
            Model pack name
        '''
        # Spacy model always should be just the name, but during loading it can be reset to path
        self.config.general['spacy_model'] = os.path.basename(self.config.general['spacy_model'])
        # Versioning
        self._versioning()
        model_pack_name += "_{}".format(self.config.version['id'])

        self.log.warning("This will save all models into a zip file, can take some time and require quite a bit of disk space.")
        _save_dir_path = save_dir_path
        save_dir_path = os.path.join(save_dir_path, model_pack_name)

        os.makedirs(save_dir_path, exist_ok=True)

        # Save the used spacy model
        spacy_path = os.path.join(save_dir_path, self.config.general['spacy_model'])
        if str(self.pipe.spacy_nlp._path) != spacy_path:
            # First remove if something is there
            shutil.rmtree(spacy_path, ignore_errors=True)
            shutil.copytree(str(self.pipe.spacy_nlp._path), spacy_path)

        # Save the CDB
        cdb_path = os.path.join(save_dir_path, "cdb.dat")
        self.cdb.save(cdb_path)

        # Save the Vocab
        vocab_path = os.path.join(save_dir_path, "vocab.dat")
        if self.vocab is None:
            raise ValueError("Model pack creation is failed due to the missing 'vocab'")
        else:
            self.vocab.save(vocab_path)

        # Save all meta_cats
        for comp in self.pipe.spacy_nlp.components:
            if isinstance(comp[1], MetaCAT):
                name = comp[0]
                meta_path = os.path.join(save_dir_path, "meta_" + name)
                comp[1].save(meta_path)

        # Add a model card also, why not
        model_card_path = os.path.join(save_dir_path, "model_card.json")
        json.dump(self.get_model_card(as_dict=True), open(model_card_path, 'w'), indent=2)

        # Zip everything
        shutil.make_archive(os.path.join(_save_dir_path, model_pack_name), 'zip', root_dir=save_dir_path)

        # Log model card and return new name
        self.log.info(self.get_model_card()) # Print the model card
        return model_pack_name

    @classmethod
    def load_model_pack(cls, zip_path: str, meta_cat_config_dict: Optional[Dict] = None) -> "CAT":
        r''' Load everything

        Args:
            zip_path
            meta_cat_config_dict:
                A config dict that will overwrite existing configs in meta_cat.
                Can be something like:
                    meta_cat_config_dict = {'general': {'device': 'cpu'}}
        '''
        from medcat.cdb import CDB
        from medcat.vocab import Vocab
        from medcat.meta_cat import MetaCAT

        base_dir = os.path.dirname(zip_path)
        filename = os.path.basename(zip_path)
        foldername = filename.replace(".zip", '')

        model_pack_path = os.path.join(base_dir, foldername)
        if os.path.exists(model_pack_path):
            cls.log.info("Found an existing unziped model pack at: {}, the provided zip will not be touched.".format(model_pack_path))
        else:
            cls.log.info("Unziping the model pack and loading models.")
            shutil.unpack_archive(zip_path, extract_dir=model_pack_path)

        # Load the CDB
        cdb_path = os.path.join(model_pack_path, "cdb.dat")
        cdb = CDB.load(cdb_path)

        # Modify the config to contain full path to spacy model
        cdb.config.general['spacy_model'] = os.path.join(model_pack_path, os.path.basename(cdb.config.general['spacy_model']))

        # Load Vocab
        vocab_path = os.path.join(model_pack_path, "vocab.dat")
        vocab = Vocab.load(vocab_path)

        # Find meta models in the model_pack
        meta_paths = [os.path.join(model_pack_path, path) for path in os.listdir(model_pack_path) if path.startswith('meta_')]
        meta_cats = []
        for meta_path in meta_paths:
            meta_cats.append(MetaCAT.load(save_dir_path=meta_path,
                                          config_dict=meta_cat_config_dict))

        cat = cls(cdb=cdb, config=cdb.config, vocab=vocab, meta_cats=meta_cats)
        cls.log.info(cat.get_model_card()) # Print the model card
        return cat

    def __call__(self, text: Optional[str], do_train: bool = False) -> Optional[Doc]:
        r'''
        Push the text through the pipeline.

        Args:
            text (string):
                The text to be annotated, if the text length is longer than
                self.config.preprocessing['max_document_length'] it will be trimmed to that length.
            do_train (bool, defaults to `False`):
                This causes so many screwups when not there, so I'll force training
                to False. To run training it is much better to use the self.train() function
                but for some special cases I'm leaving it here also.
        Returns:
            A single spacy document or multiple spacy documents with the extracted entities
        '''
        # Should we train - do not use this for training, unless you know what you are doing. Use the
        #self.train() function
        self.config.linking['train'] = do_train

        if text is None:
            self.log.error("The input text should be either a string or a sequence of strings but got %s", type(text))
            return None
        else:
            text = self._get_trimmed_text(str(text))
            return self.pipe(text)

    def _print_stats(self,
                     data: Dict,
                     epoch: int = 0,
                     use_project_filters: bool = False,
                     use_overlaps: bool = False,
                     use_cui_doc_limit: bool = False,
                     use_groups: bool = False,
                     extra_cui_filter: Optional[Set] = None) -> Tuple:
        r''' TODO: Refactor and make nice
        Print metrics on a dataset (F1, P, R), it will also print the concepts that have the most FP,FN,TP.

        Args:
            data (list of dict):
                The json object that we get from MedCATtrainer on export.
            epoch (int):
                Used during training, so we know what epoch is it.
            use_project_filters (boolean):
                Each project in medcattrainer can have filters, do we want to respect those filters
                when calculating metrics.
            use_overlaps (boolean):
                Allow overlapping entites, nearly always False as it is very difficult to annotate overlapping entites.
            use_cui_doc_limit (boolean):
                If True the metrics for a CUI will be only calculated if that CUI appears in a document, in other words
                if the document was annotated for that CUI. Useful in very specific situations when during the annotation
                process the set of CUIs changed.
            use_groups (boolean):
                If True concepts that have groups will be combined and stats will be reported on groups.
            extra_cui_filter(Optional[Set]):
                This filter will be intersected with all other filters, or if all others are not set then only this one will be used.

        Returns:
            fps (dict):
                False positives for each CUI
            fns (dict):
                False negatives for each CUI
            tps (dict):
                True positives for each CUI
            cui_prec (dict):
                Precision for each CUI
            cui_rec (dict):
                Recall for each CUI
            cui_f1 (dict):
                F1 for each CUI
            cui_counts (dict):
                Number of occurrence for each CUI
            examples (dict):
                Examples for each of the fp, fn, tp. Format will be examples['fp']['cui'][<list_of_examples>]
        '''
        tp = 0
        fp = 0
        fn = 0
        fps: Dict = {}
        fns: Dict = {}
        tps: Dict = {}
        cui_prec: Dict = {}
        cui_rec: Dict = {}
        cui_f1: Dict = {}
        cui_counts: Dict = {}
        examples: Dict = {'fp': {}, 'fn': {}, 'tp': {}}

        fp_docs: Set = set()
        fn_docs: Set = set()
        # reset and back up filters
        _filters = deepcopy(self.config.linking['filters'])
        filters = self.config.linking['filters']
        for pind, project in tqdm(enumerate(data['projects']), desc="Stats project", total=len(data['projects']), leave=False):
            filters['cuis'] = set()

            # Add extrafilter if set
            if isinstance(extra_cui_filter, set):
                filters['cuis'] = extra_cui_filter

            if use_project_filters:
                project_filter = get_project_filters(cuis=project.get('cuis', None),
                                                      type_ids=project.get('tuis', None),
                                                      cdb=self.cdb,
                                                      project=project)
                # Intersect project filter with existing if it has something
                if project_filter:
                    filters['cuis'] = intersect_nonempty_set(project_filter, filters['cuis'])

            for dind, doc in tqdm(
                enumerate(project["documents"]),
                desc="Stats document",
                total=len(project["documents"]),
                leave=False,
            ):
                anns = self._get_doc_annotations(doc)

                # Apply document level filtering, in this case project_filter is ignored while the extra_cui_filter is respected still
                if use_cui_doc_limit:
                    _cuis = set([ann['cui'] for ann in anns])
                    if _cuis:
                        filters['cuis'] = intersect_nonempty_set(_cuis, extra_cui_filter)
                    else:
                        filters['cuis'] = {'empty'}

                spacy_doc: Doc = self(doc['text'])

                if use_overlaps:
                    p_anns = spacy_doc._.ents
                else:
                    p_anns = spacy_doc.ents

                anns_norm = []
                anns_norm_neg = []
                anns_examples = []
                anns_norm_cui = []
                for ann in anns:
                    cui = ann['cui']
                    if check_filters(cui, filters):
                        if use_groups:
                            cui = self.cdb.addl_info['cui2group'].get(cui, cui)

                        if ann.get('validated', True) and (not ann.get('killed', False) and not ann.get('deleted', False)):
                            anns_norm.append((ann['start'], cui))
                            anns_examples.append({"text": doc['text'][max(0, ann['start']-60):ann['end']+60],
                                                  "cui": cui,
                                                  "source value": ann['value'],
                                                  "acc": 1,
                                                  "project index": pind,
                                                  "document inedex": dind})
                        elif ann.get('validated', True) and (ann.get('killed', False) or ann.get('deleted', False)):
                            anns_norm_neg.append((ann['start'], cui))


                        if ann.get("validated", True):
                            # This is used to test was someone annotating for this CUI in this document
                            anns_norm_cui.append(cui)
                            cui_counts[cui] = cui_counts.get(cui, 0) + 1

                p_anns_norm = []
                p_anns_examples = []
                for ann in p_anns:
                    cui = ann._.cui
                    if use_groups:
                        cui = self.cdb.addl_info['cui2group'].get(cui, cui)

                    p_anns_norm.append((ann.start_char, cui))
                    p_anns_examples.append({"text": doc['text'][max(0, ann.start_char-60):ann.end_char+60],
                                          "cui": cui,
                                          "source value": ann.text,
                                          "acc": float(ann._.context_similarity),
                                          "project index": pind,
                                          "document inedex": dind})


                for iann, ann in enumerate(p_anns_norm):
                    cui = ann[1]
                    if ann in anns_norm:
                        tp += 1
                        tps[cui] = tps.get(cui, 0) + 1

                        example = p_anns_examples[iann]
                        examples['tp'][cui] = examples['tp'].get(cui, []) + [example]
                    else:
                        fp += 1
                        fps[cui] = fps.get(cui, 0) + 1
                        fp_docs.add(doc.get('name', 'unk'))

                        # Add example for this FP prediction
                        example = p_anns_examples[iann]
                        if ann in anns_norm_neg:
                            # Means that it really was annotated as negative
                            example['real_fp'] = True

                        examples['fp'][cui] = examples['fp'].get(cui, []) + [example]

                for iann, ann in enumerate(anns_norm):
                    if ann not in p_anns_norm:
                        cui = ann[1]
                        fn += 1
                        fn_docs.add(doc.get('name', 'unk'))

                        fns[cui] = fns.get(cui, 0) + 1
                        examples['fn'][cui] = examples['fn'].get(cui, []) + [anns_examples[iann]]

        try:
            prec = tp / (tp + fp)
            rec = tp / (tp + fn)
            f1 = 2*(prec*rec) / (prec + rec)
            print("Epoch: {}, Prec: {}, Rec: {}, F1: {}\n".format(epoch, prec, rec, f1))
            print("Docs with false positives: {}\n".format("; ".join([str(x) for x in list(fp_docs)[0:10]])))
            print("Docs with false negatives: {}\n".format("; ".join([str(x) for x in list(fn_docs)[0:10]])))

            # Sort fns & prec
            fps = {k: v for k, v in sorted(fps.items(), key=lambda item: item[1], reverse=True)}
            fns = {k: v for k, v in sorted(fns.items(), key=lambda item: item[1], reverse=True)}
            tps = {k: v for k, v in sorted(tps.items(), key=lambda item: item[1], reverse=True)}


            # F1 per concept
            for cui in tps.keys():
                prec = tps[cui] / (tps.get(cui, 0) + fps.get(cui, 0))
                rec = tps[cui] / (tps.get(cui, 0) + fns.get(cui, 0))
                f1 = 2*(prec*rec) / (prec + rec)
                cui_prec[cui] = prec
                cui_rec[cui] = rec
                cui_f1[cui] = f1


            # Get top 10
            pr_fps = [(self.cdb.cui2preferred_name.get(cui,
                list(self.cdb.cui2names.get(cui, [cui]))[0]), cui, fps[cui]) for cui in list(fps.keys())[0:10]]
            pr_fns = [(self.cdb.cui2preferred_name.get(cui,
                list(self.cdb.cui2names.get(cui, [cui]))[0]), cui, fns[cui]) for cui in list(fns.keys())[0:10]]
            pr_tps = [(self.cdb.cui2preferred_name.get(cui,
                list(self.cdb.cui2names.get(cui, [cui]))[0]), cui, tps[cui]) for cui in list(tps.keys())[0:10]]


            print("\n\nFalse Positives\n")
            for one in pr_fps:
                print("{:70} - {:20} - {:10}".format(str(one[0])[0:69], str(one[1])[0:19], one[2]))
            print("\n\nFalse Negatives\n")
            for one in pr_fns:
                print("{:70} - {:20} - {:10}".format(str(one[0])[0:69], str(one[1])[0:19], one[2]))
            print("\n\nTrue Positives\n")
            for one in pr_tps:
                print("{:70} - {:20} - {:10}".format(str(one[0])[0:69], str(one[1])[0:19], one[2]))
            print("*"*110 + "\n")

        except Exception:
            traceback.print_exc()

        # restore filters to original state
        self.config.linking['filters'] = _filters

        return fps, fns, tps, cui_prec, cui_rec, cui_f1, cui_counts, examples

    def _init_ckpts(self, is_resumed, checkpoint):
        if self.config.general['checkpoint']['steps'] is not None or checkpoint is not None:
            checkpoint_config = CheckpointConfig(**self.config.general.get('checkpoint', {}))
            checkpoint_manager = CheckpointManager('cat_train', checkpoint_config)
            if is_resumed:
                # TODO: probably remove is_resumed mark and always resume if a checkpoint is provided,
                #but I'll leave it for now
                checkpoint = checkpoint or checkpoint_manager.get_latest_checkpoint()
                self.log.info(f"Resume training on the most recent checkpoint at {checkpoint.dir_path}...")
                self.cdb = checkpoint.restore_latest_cdb()
                self.cdb.config.merge_config(self.config.__dict__)
                self.config = self.cdb.config
                self._create_pipeline(self.config)
            else:
                checkpoint = checkpoint or checkpoint_manager.create_checkpoint()
                self.log.info(f"Start new training and checkpoints will be saved at {checkpoint.dir_path}...")

        return checkpoint

    def train(self,
              data_iterator: Iterable,
              nepochs: int = 1,
              fine_tune: bool = True,
              progress_print: int = 1000,
              checkpoint: Optional[Checkpoint] = None,
              is_resumed: bool = False) -> None:
        """ Runs training on the data, note that the maximum length of a line
        or document is 1M characters. Anything longer will be trimmed.

        Args:
            data_iterator (Iterable):
                Simple iterator over sentences/documents, e.g. a open file
                or an array or anything that we can use in a for loop.
            nepochs (int):
                Number of epochs for which to run the training.
            fine_tune (bool):
                If False old training will be removed.
            progress_print (int):
                Print progress after N lines.
            checkpoint (Optional[medcat.utils.checkpoint.CheckpointUT]):
                The MedCAT checkpoint object
            is_resumed (bool):
                If True resume the previous training; If False, start a fresh new training.
        """
        if not fine_tune:
            self.log.info("Removing old training data!")
            self.cdb.reset_training()
        checkpoint = self._init_ckpts(is_resumed, checkpoint)

        latest_trained_step = checkpoint.count if checkpoint is not None else 0
        epochal_data_iterator = chain.from_iterable(repeat(data_iterator, nepochs))
        for line in islice(epochal_data_iterator, latest_trained_step, None):
            if line is not None and line:
                # Convert to string
                line = str(line).strip()

                try:
                    _ = self(line, do_train=True)
                except Exception as e:
                    self.log.warning("LINE: '%s...' \t WAS SKIPPED", line[0:100])
                    self.log.warning("BECAUSE OF: %s", str(e))
            else:
                self.log.warning("EMPTY LINE WAS DETECTED AND SKIPPED")

            latest_trained_step += 1
            if latest_trained_step % progress_print == 0:
                self.log.info("DONE: %s", str(latest_trained_step))
            if checkpoint is not None and checkpoint.steps is not None and latest_trained_step % checkpoint.steps == 0:
                checkpoint.save(cdb=self.cdb, count=latest_trained_step)

        self.config.linking['train'] = False

    def add_cui_to_group(self, cui: str, group_name: str) -> None:
        r'''
        Ads a CUI to a group, will appear in cdb.addl_info['cui2group']

        Args:
            cui (str):
                The concept to be added
            group_name (str):
                The group to whcih the concept will be added

        Examples:
            >>> cat.add_cui_to_group("S-17", 'pain')
        '''

        # Add group_name
        self.cdb.addl_info['cui2group'][cui] = group_name

    def unlink_concept_name(self, cui: str, name: str, preprocessed_name: bool = False) -> None:
        r'''
        Unlink a concept name from the CUI (or all CUIs if full_unlink), removes the link from
        the Concept Database (CDB). As a consequence medcat will never again link the `name`
        to this CUI - meaning the name will not be detected as a concept in the future.

        Args:
            cui (str):
                The CUI from which the `name` will be removed
            name (str):
                The span of text to be removed from the linking dictionary
        Examples:
            >>> # To never again link C0020538 to HTN
            >>> cat.unlink_concept_name('C0020538', 'htn', False)
        '''

        cuis = [cui]
        if preprocessed_name:
            names = {name: 'nothing'}
        else:
            names = prepare_name(name, self.pipe.spacy_nlp, {}, self.config)

        # If full unlink find all CUIs
        if self.config.general.get('full_unlink', False):
            for n in names:
                cuis.extend(self.cdb.name2cuis.get(n, []))

        # Remove name from all CUIs
        for c in cuis:
            self.cdb.remove_names(cui=c, names=names)

    def add_and_train_concept(self,
                              cui: str,
                              name: str,
                              spacy_doc: Optional[Doc] = None,
                              spacy_entity: Optional[Union[List[Token], Span]] = None,
                              ontologies: Set = set(),
                              name_status: str = 'A',
                              type_ids: Set = set(),
                              description: str = '',
                              full_build: bool = True,
                              negative: bool = False,
                              devalue_others: bool = False,
                              do_add_concept: bool = True) -> None:
        r''' Add a name to an existing concept, or add a new concept, or do not do anything if the name or concept already exists. Perform
        training if spacy_entity and spacy_doc are set.

        Args:
            cui (str):
                CUI of the concept
            name (str):
                Name to be linked to the concept (in the case of MedCATtrainer this is simply the
                selected value in text, no preprocessing or anything needed).
            spacy_doc (spacy.tokens.Doc):
                Spacy represenation of the document that was manually annotated.
            spacy_entity (Optional[Union[List[Token], Span]]):
                Given the spacy document, this is the annotated span of text - list of annotated tokens that are marked with this CUI.
            negative (bool):
                Is this a negative or positive example.
            devalue_others:
                If set, cuis to which this name is assigned and are not `cui` will receive negative training given
                that negative=False.

            **other:
                Refer to CDB.add_concept
        '''
        names = prepare_name(name, self.pipe.spacy_nlp, {}, self.config)
        # Only if not negative, otherwise do not add the new name if in fact it should not be detected
        if do_add_concept and not negative:
            self.cdb.add_concept(cui=cui, names=names, ontologies=ontologies, name_status=name_status, type_ids=type_ids, description=description,
                                 full_build=full_build)

        if spacy_entity is not None and spacy_doc is not None:
            # Train Linking
            self.linker.context_model.train(cui=cui, entity=spacy_entity, doc=spacy_doc, negative=negative, names=names)

            if not negative and devalue_others:
                # Find all cuis
                cuis = set()
                for n in names:
                    cuis.update(self.cdb.name2cuis.get(n, []))
                # Remove the cui for which we just added positive training
                if cui in cuis:
                    cuis.remove(cui)
                # Add negative training for all other CUIs that link to these names
                for _cui in cuis:
                    self.linker.context_model.train(cui=_cui, entity=spacy_entity, doc=spacy_doc, negative=True)

    def train_supervised(self,
                         data_path: str,
                         reset_cui_count: bool = False,
                         nepochs: int = 1,
                         print_stats: int = 0,
                         use_filters: bool = False,
                         terminate_last: bool = False,
                         use_overlaps: bool = False,
                         use_cui_doc_limit: bool = False,
                         test_size: int = 0,
                         devalue_others: bool = False,
                         use_groups: bool = False,
                         never_terminate: bool = False,
                         train_from_false_positives: bool = False,
                         extra_cui_filter: Optional[Set] = None,
                         checkpoint: Optional[Checkpoint] = None,
                         is_resumed: bool = False) -> Tuple:
        r''' TODO: Refactor, left from old
        Run supervised training on a dataset from MedCATtrainer. Please take care that this is more a simulated
        online training then supervised.

        Args:
            data_path (str):
                The path to the json file that we get from MedCATtrainer on export.
            reset_cui_count (boolean):
                Used for training with weight_decay (annealing). Each concept has a count that is there
                from the beginning of the CDB, that count is used for annealing. Resetting the count will
                significantly increase the training impact. This will reset the count only for concepts
                that exist in the the training data.
            nepochs (int):
                Number of epochs for which to run the training.
            print_stats (int):
                If > 0 it will print stats every print_stats epochs.
            use_filters (boolean):
                Each project in medcattrainer can have filters, do we want to respect those filters
                when calculating metrics.
            terminate_last (boolean):
                If true, concept termination will be done after all training.
            use_overlaps (boolean):
                Allow overlapping entities, nearly always False as it is very difficult to annotate overlapping entities.
            use_cui_doc_limit (boolean):
                If True the metrics for a CUI will be only calculated if that CUI appears in a document, in other words
                if the document was annotated for that CUI. Useful in very specific situations when during the annotation
                process the set of CUIs changed.
            test_size (float):
                If > 0 the data set will be split into train test based on this ration. Should be between 0 and 1.
                Usually 0.1 is fine.
            devalue_others(bool):
                Check add_name for more details.
            use_groups (boolean):
                If True concepts that have groups will be combined and stats will be reported on groups.
            never_terminate (boolean):
                If True no termination will be applied
            train_from_false_positives (boolean):
                If True it will use false positive examples detected by medcat and train from them as negative examples.
            extra_cui_filter(Optional[Set]):
                This filter will be intersected with all other filters, or if all others are not set then only this one will be used.
            checkpoint (Optional[Optional[medcat.utils.checkpoint.CheckpointST]):
                The MedCAT CheckpointST object
            is_resumed (bool):
                If True resume the previous training; If False, start a fresh new training.
        Returns:
            fp (dict):
                False positives for each CUI
            fn (dict):
                False negatives for each CUI
            tp (dict):
                True positives for each CUI
            p (dict):
                Precision for each CUI
            r (dict):
                Recall for each CUI
            f1 (dict):
                F1 for each CUI
            cui_counts (dict):
                Number of occurrence for each CUI
            examples (dict):
                FP/FN examples of sentences for each CUI
        '''
        checkpoint = self._init_ckpts(is_resumed, checkpoint)

        # Backup filters
        _filters = deepcopy(self.config.linking['filters'])
        filters = self.config.linking['filters']

        fp = fn = tp = p = r = f1 = examples = {}
        with open(data_path) as f:
            data = json.load(f)
        cui_counts = {}

        if test_size == 0:
            self.log.info("Running without a test set, or train==test")
            test_set = data
            train_set = data
        else:
            train_set, test_set, _, _ = make_mc_train_test(data, self.cdb, test_size=test_size)

        if print_stats > 0:
            fp, fn, tp, p, r, f1, cui_counts, examples = self._print_stats(test_set,
                                                                           use_project_filters=use_filters,
                                                                           use_cui_doc_limit=use_cui_doc_limit,
                                                                           use_overlaps=use_overlaps,
                                                                           use_groups=use_groups,
                                                                           extra_cui_filter=extra_cui_filter)
        if reset_cui_count:
            # Get all CUIs
            cuis = []
            for project in train_set['projects']:
                for doc in project['documents']:
                    doc_annotations = self._get_doc_annotations(doc)
                    for ann in doc_annotations:
                        cuis.append(ann['cui'])
            for cui in set(cuis):
                if cui in self.cdb.cui2count_train:
                    self.cdb.cui2count_train[cui] = 10

        # Remove entities that were terminated
        if not never_terminate:
            for project in train_set['projects']:
                for doc in project['documents']:
                    doc_annotations = self._get_doc_annotations(doc)
                    for ann in doc_annotations:
                        if ann.get('killed', False):
                            self.unlink_concept_name(ann['cui'], ann['value'])

        latest_trained_step = checkpoint.count if checkpoint is not None else 0
        current_epoch, current_project, current_document = self._get_training_start(train_set, latest_trained_step)

        for epoch in trange(current_epoch, nepochs, initial=current_epoch, total=nepochs, desc='Epoch', leave=False):
            # Print acc before training
            for idx_project in trange(current_project, len(train_set['projects']), initial=current_project, total=len(train_set['projects']), desc='Project', leave=False):
                project = train_set['projects'][idx_project]

                # Set filters in case we are using the train_from_fp
                filters['cuis'] = set()
                if isinstance(extra_cui_filter, set):
                    filters['cuis'] = extra_cui_filter

                if use_filters:
                    project_filter = get_project_filters(cuis=project.get('cuis', None),
                                                         type_ids=project.get('tuis', None),
                                                         cdb=self.cdb,
                                                         project=project)

                    if project_filter:
                        filters['cuis'] = intersect_nonempty_set(project_filter, filters['cuis'])

                for idx_doc in trange(current_document, len(project['documents']), initial=current_document, total=len(project['documents']), desc='Document', leave=False):
                    doc = project['documents'][idx_doc]
                    spacy_doc: Doc = self(doc['text'])

                    # Compatibility with old output where annotations are a list
                    doc_annotations = self._get_doc_annotations(doc)
                    for ann in doc_annotations:
                        if not ann.get('killed', False):
                            cui = ann['cui']
                            start = ann['start']
                            end = ann['end']
                            spacy_entity = tkns_from_doc(spacy_doc=spacy_doc, start=start, end=end)
                            deleted = ann.get('deleted', False)
                            self.add_and_train_concept(cui=cui,
                                                       name=ann['value'],
                                                       spacy_doc=spacy_doc,
                                                       spacy_entity=spacy_entity,
                                                       negative=deleted,
                                                       devalue_others=devalue_others)
                    if train_from_false_positives:
                        fps: List[Span] = get_false_positives(doc, spacy_doc)

                        for fp in fps:
                            fp_: Span = fp
                            self.add_and_train_concept(cui=fp_._.cui,
                                                       name=fp_.text,
                                                       spacy_doc=spacy_doc,
                                                       spacy_entity=fp_,
                                                       negative=True,
                                                       do_add_concept=False)

                    latest_trained_step += 1
                    if checkpoint is not None and checkpoint.steps is not None and latest_trained_step % checkpoint.steps == 0:
                        checkpoint.save(self.cdb, latest_trained_step)

            if terminate_last and not never_terminate:
                # Remove entities that were terminated, but after all training is done
                for project in train_set['projects']:
                    for doc in project['documents']:
                        doc_annotations = self._get_doc_annotations(doc)
                        for ann in doc_annotations:
                            if ann.get('killed', False):
                                self.unlink_concept_name(ann['cui'], ann['value'])

            if print_stats > 0 and (epoch + 1) % print_stats == 0:
                fp, fn, tp, p, r, f1, cui_counts, examples = self._print_stats(test_set,
                                                                               epoch=epoch + 1,
                                                                               use_project_filters=use_filters,
                                                                               use_cui_doc_limit=use_cui_doc_limit,
                                                                               use_overlaps=use_overlaps,
                                                                               use_groups=use_groups,
                                                                               extra_cui_filter=extra_cui_filter)

        # Set the filters again
        self.config.linking['filters'] = _filters

        return fp, fn, tp, p, r, f1, cui_counts, examples

    def get_entities(self,
                     text: str,
                     only_cui: bool = False,
                     addl_info: List[str] = ['cui2icd10', 'cui2ontologies', 'cui2snomed']) -> Dict:
        doc = self(text)
        out = self._doc_to_out(doc, only_cui, addl_info)
        return out

    def get_entities_multi_texts(self,
                     texts: Union[Iterable[str], Iterable[Tuple]],
                     only_cui: bool = False,
                     addl_info: List[str] = ['cui2icd10', 'cui2ontologies', 'cui2snomed'],
                     n_process: Optional[int] = None,
                     batch_size: Optional[int] = None) -> List[Dict]:
        r''' Get entities
        text:  text to be annotated
        return:  entities
        '''
        out: List[Dict] = []

        if n_process is None:
            texts_ = self._generate_trimmed_texts(texts)
            for text in texts_:
                out.append(self._doc_to_out(self(text), only_cui, addl_info))
        else:
            self.pipe.set_error_handler(self._pipe_error_handler)
            try:
                texts_ = self._get_trimmed_texts(texts)
                docs = self.pipe.batch_multi_process(texts_, n_process, batch_size)

                for doc in tqdm(docs, total=len(texts_)):
                    doc = None if doc.text.strip() == '' else doc
                    out.append(self._doc_to_out(doc, only_cui, addl_info, out_with_text=True))

                # Currently spaCy cannot mark which pieces of texts failed within the pipe so be this workaround,
                # which also assumes texts are different from each others.
                if len(out) < len(texts_):
                    self.log.warning("Found at least one failed batch and set output for enclosed texts to empty")
                    for i, text in enumerate(texts_):
                        if i == len(out):
                            out.append(self._doc_to_out(None, only_cui, addl_info))
                        elif out[i].get('text', '') != text:
                            out.insert(i, self._doc_to_out(None, only_cui, addl_info))

                cnf_annotation_output = getattr(self.config, 'annotation_output', {})
                if not(cnf_annotation_output.get('include_text_in_output', False)):
                    for o in out:
                        if o is not None:
                            o.pop('text', None)
            finally:
                self.pipe.reset_error_handler()

        return out

    def get_json(self, text: str, only_cui: bool = False, addl_info=['cui2icd10', 'cui2ontologies']) -> str:
        """ Get output in json format

        text:  text to be annotated
        return:  json with fields {'entities': <>, 'text': text}
        """
        ents = self.get_entities(text, only_cui, addl_info=addl_info)['entities']
        out = {'annotations': ents, 'text': text}

        return json.dumps(out)

    @staticmethod
    def _get_training_start(train_set, latest_trained_step):
        total_steps_per_epoch = sum([1 for project in train_set['projects'] for _ in project['documents']])
        if total_steps_per_epoch == 0:
            raise ValueError("MedCATtrainer export contains no documents")
        current_epoch, last_step_in_epoch = divmod(latest_trained_step, total_steps_per_epoch)
        document_count = 0
        current_project = 0
        current_document = 0
        for idx_project, project in enumerate(train_set['projects']):
            for idx_doc, _ in enumerate(project['documents']):
                document_count += 1
                if document_count == last_step_in_epoch:
                    current_project = idx_project
                    current_document = idx_doc
                    break
            if current_project > 0:
                break
            current_document = 0
        return current_epoch, current_project, current_document

    def _separate_nn_components(self):
        # Loop though the models and check are there GPU devices
        nn_components = []
        for component in self.pipe.spacy_nlp.components:
            if isinstance(component[1], MetaCAT):
                self.pipe.spacy_nlp.disable_pipe(component[0])
                nn_components.append(component)

        return nn_components

    def _run_nn_components(self, docs: Dict, nn_components: List, id2text: Dict) -> None:
        r''' This will add meta_anns in-place to the docs dict.
        '''
        self.log.debug("Running GPU components separately")

        # First convert the docs into the fake spacy doc format
        spacy_docs = json_to_fake_spacy(docs, id2text=id2text)
        # Disable component locks also
        for name, component in nn_components:
            component.config.general['disable_component_lock'] = True

        for name, component in nn_components:
            spacy_docs = component.pipe(spacy_docs)

        for spacy_doc in spacy_docs:
            for ent in spacy_doc.ents:
                docs[spacy_doc.id]['entities'][ent._.id]['meta_anns'].update(ent._.meta_anns)

    def _batch_generator(self, data: Iterable, batch_size_chars: int, skip_ids: Set = set()):
        docs = []
        char_count = 0
        for doc in data:
            if doc[0] not in skip_ids:
                char_count += len(str(doc[1]))
                docs.append(doc)
                if char_count < batch_size_chars:
                    continue
                yield docs
                docs = []
                char_count = 0

        if len(docs) > 0:
            yield docs

    def _save_docs_to_file(self, docs: Iterable, annotated_ids: List[str], save_dir_path: str, annotated_ids_path: Optional[str], part_counter: int = 0) -> int:
        path = os.path.join(save_dir_path, 'part_{}.pickle'.format(part_counter))
        pickle.dump(docs, open(path, "wb"))
        self.log.info("Saved part: %s, to: %s", part_counter, path)
        part_counter = part_counter + 1 # Increase for save, as it should be what is the next part
        if annotated_ids_path is not None:
            pickle.dump((annotated_ids, part_counter), open(annotated_ids_path, 'wb'))
        return part_counter

    def multiprocessing(self,
                        data: Union[List[Tuple], Iterable[Tuple]],
                        nproc: int = 2,
                        batch_size_chars: int = 5000 * 1000,
                        only_cui: bool = False,
                        addl_info: List[str] = [],
                        separate_nn_components: bool = True,
                        out_split_size_chars: Optional[int] = None,
                        save_dir_path: str = os.path.abspath(os.getcwd()),
                        min_free_memory=0.1) -> Dict:
        r''' Run multiprocessing for inference, if out_save_path and out_split_size_chars is used this will also continue annotating
        documents if something is saved in that directory.

        Args:
            data(``):
                Iterator or array with format: [(id, text), (id, text), ...]
            nproc (`int`, defaults to 8):
                Number of processors
            batch_size_chars (`int`, defaults to 1000000):
                Size of a batch in number of characters, this should be around: NPROC * average_document_length * 200
            separate_nn_components (`bool`, defaults to True):
                If set the medcat pipe will be broken up into NN and not-NN components and
                they will be run sequentially. This is useful as the NN components
                have batching and like to process many docs at once, while the rest of the pipeline
                runs the documents one by one.
            out_split_size_chars (`int`, None):
                If set once more than out_split_size_chars are annotated
                they will be saved to a file (save_dir_path) and the memory cleared. Recommended
                value is 20*batch_size_chars.
            save_dir_path(`str`, defaults to the current working directory):
                Where to save the annotated documents if splitting.
            min_free_memory(`float`, defaults to 0.1):
                If set a process will not start unless there is at least this much RAM memory left,
                should be a range between [0, 1] meaning how much of the memory has to be free. Helps when annotating
                very large datasets because spacy is not the best with memory management and multiprocessing.

        Returns:
            A dictionary: {id: doc_json, id2: doc_json2, ...}, in case out_split_size_chars is used
            the last batch will be returned while that and all previous batches will be
            written to disk (out_save_dir).
        '''
        # Set max document length
        self.pipe.spacy_nlp.max_length = self.config.preprocessing.get('max_document_length', 1000000)

        if self._meta_cats and not separate_nn_components:
            # Hack for torch using multithreading, which is not good if not 
            #separate_nn_components, need for CPU runs only
            import torch
            torch.set_num_threads(1)

        nn_components = []
        if separate_nn_components:
            nn_components = self._separate_nn_components()

        if save_dir_path is not None:
            os.makedirs(save_dir_path, exist_ok=True)

        # "5" looks like a magic number here so better with comment about why the choice was made.
        internal_batch_size_chars = batch_size_chars // (5 * nproc)

        annotated_ids_path = os.path.join(save_dir_path, 'annotated_ids.pickle') if save_dir_path is not None else None
        if annotated_ids_path is not None and os.path.exists(annotated_ids_path):
            annotated_ids, part_counter = pickle.load(open(annotated_ids_path, 'rb'))
        else:
            annotated_ids = []
            part_counter = 0

        docs = {}
        _start_time = time.time()
        _batch_counter = 0 # Used for splitting the output, counts batches inbetween saves
        for batch in self._batch_generator(data, batch_size_chars, skip_ids=set(annotated_ids)):
            self.log.info("Annotated until now: %s docs; Current BS: %s docs; Elapsed time: %.2f minutes",
                          len(annotated_ids),
                          len(batch),
                          (time.time() - _start_time)/60)
            try:
                _docs = self._multiprocessing_batch(data=batch,
                                                    nproc=nproc,
                                                    only_cui=only_cui,
                                                    batch_size_chars=internal_batch_size_chars,
                                                    addl_info=addl_info,
                                                    nn_components=nn_components,
                                                    min_free_memory=min_free_memory)
                docs.update(_docs)
                annotated_ids.extend(_docs.keys())
                _batch_counter += 1
                del _docs
                if out_split_size_chars is not None and (_batch_counter * batch_size_chars) > out_split_size_chars:
                    # Save to file and reset the docs 
                    part_counter = self._save_docs_to_file(docs=docs,
                                           annotated_ids=annotated_ids,
                                           save_dir_path=save_dir_path,
                                           annotated_ids_path=annotated_ids_path,
                                           part_counter=part_counter)
                    del docs
                    docs = {}
                    _batch_counter = 0
            except Exception as e:
                self.log.warning("Failed an outer batch in the multiprocessing script")
                self.log.warning(e, exc_info=True, stack_info=True)

        # Save the last batch
        if out_split_size_chars is not None and len(docs) > 0:
            # Save to file and reset the docs 
            self._save_docs_to_file(docs=docs,
                                   annotated_ids=annotated_ids,
                                   save_dir_path=save_dir_path,
                                   annotated_ids_path=annotated_ids_path,
                                   part_counter=part_counter)

        # Enable the GPU Components again
        if separate_nn_components:
            for name, _ in nn_components:
                # No need to do anything else as it was already in the pipe
                self.pipe.spacy_nlp.enable_pipe(name)

        return docs

    def _multiprocessing_batch(self,
                               data: Union[List[Tuple], Iterable[Tuple]],
                               nproc: int = 8,
                               batch_size_chars: int = 1000000,
                               only_cui: bool = False,
                               addl_info: List[str] = [],
                               nn_components: List = [],
                               min_free_memory: int = 0) -> Dict:
        r''' Run multiprocessing on one batch

        Args:
            data(``):
                Iterator or array with format: [(id, text), (id, text), ...]
            nproc (`int`, defaults to 8):
                Number of processors
            batch_size_chars (`int`, defaults to 1000000):
                Size of a batch in number of characters

        Returns:
            A dictionary: {id: doc_json, id2: doc_json2, ...}
        '''
        # Create the input output for MP
        with Manager() as manager:
            out_list = manager.list()
            lock = manager.Lock()
            in_q = manager.Queue(maxsize=10*nproc)

            id2text = {}
            for batch in self._batch_generator(data, batch_size_chars):
                if nn_components:
                    # We need this for the json_to_fake_spacy
                    id2text.update({k:v for k,v in batch})
                in_q.put(batch)

            # Final data point for workers
            for _ in range(nproc):
                in_q.put(None)
            sleep(2)

            # Create processes
            procs = []
            for i in range(nproc):
                p = Process(target=self._mp_cons,
                            kwargs={'in_q': in_q,
                                    'out_list': out_list,
                                    'pid': i,
                                    'only_cui': only_cui,
                                    'addl_info': addl_info,
                                    'min_free_memory': min_free_memory,
                                    'lock': lock})
                p.start()
                procs.append(p)

            # Join processes
            for p in procs:
                p.join()

            docs = {}
            # Covnerts a touple into a dict
            docs.update({k:v for k,v in out_list})

        # If we have separate GPU components now we pipe that
        if nn_components:
            try:
                self._run_nn_components(docs, nn_components, id2text=id2text)
            except Exception as e:
                self.log.warning(e, exc_info=True, stack_info=True)

        return docs

    def multiprocessing_pipe(self,
                             in_data: Union[List[Tuple], Iterable[Tuple]],
                             nproc: Optional[int] = None,
                             batch_size: Optional[int] = None,
                             only_cui: bool = False,
                             addl_info: List[str] = [],
                             return_dict: bool = True,
                             batch_factor: int = 2) -> Union[List[Tuple], Dict]:
        r''' Run multiprocessing NOT FOR TRAINING

        in_data:  a list with format: [(id, text), (id, text), ...]
        nproc:  the number of processors
        batch_size: the number of texts to buffer
        return_dict: a flag for returning either a dict or a list of tuples

        return:  a dict: {id: doc_json, id: doc_json, ...} or if return_dict is False, a list of tuples: [(id, doc_json), (id, doc_json), ...]
        '''
        out: Union[Dict, List[Tuple]]

        if nproc == 0:
            raise ValueError("nproc cannot be set to zero")

        in_data = list(in_data) if isinstance(in_data, Iterable) else in_data
        n_process = nproc if nproc is not None else min(max(cpu_count() - 1, 1), math.ceil(len(in_data) / batch_factor))
        batch_size = batch_size if batch_size is not None else math.ceil(len(in_data) / (batch_factor * abs(n_process)))

        start_method = None
        try:
            if self._meta_cats:
                import torch
                if torch.multiprocessing.get_start_method() != "spawn":
                    start_method = torch.multiprocessing.get_start_method()
                    torch.multiprocessing.set_start_method("spawn", force=True)

            entities = self.get_entities_multi_texts(texts=in_data, only_cui=only_cui, addl_info=addl_info,
                                                     n_process=n_process, batch_size=batch_size)
        finally:
            if start_method is not None:
                import torch
                torch.multiprocessing.set_start_method(start_method, force=True)

        if return_dict:
            out = {}
            for idx, data in enumerate(in_data):
                out[data[0]] = entities[idx]
        else:
            out = []
            for idx, data in enumerate(in_data):
                out.append((data[0], entities[idx]))

        return out

    def _mp_cons(self, in_q: Queue, out_list: List, min_free_memory: int, lock: Lock, pid: int = 0, only_cui: bool = False, addl_info: List = []) -> None:
        out: List = []

        while True:
            if not in_q.empty():
                if psutil.virtual_memory().available / psutil.virtual_memory().total < min_free_memory:
                    with lock:
                        out_list.extend(out)
                    # Stop a process if there is not enough memory left
                    break

                data = in_q.get()
                if data is None:
                    with lock:
                        out_list.extend(out)
                    break

                for i_text, text in data:
                    try:
                        # Annotate document
                        doc = self.get_entities(text=text, only_cui=only_cui, addl_info=addl_info)
                        out.append((i_text, doc))
                    except Exception as e:
                        self.log.warning("PID: %s failed one document in _mp_cons, running will continue normally. \n" +
                                         "Document length in chars: %s, and ID: %s", pid, len(str(text)), i_text)
                        self.log.warning(str(e))
        sleep(2)

    def _doc_to_out(self,
                    doc: Doc,
                    only_cui: bool,
                    addl_info: List[str],
                    out_with_text: bool = False) -> Dict:
        out: Dict = {'entities': {}, 'tokens': []}
        cnf_annotation_output = getattr(self.config, 'annotation_output', {})
        if doc is not None:
            out_ent: Dict = {}
            if self.config.general.get('show_nested_entities', False):
                _ents = []
                for _ent in doc._.ents:
                    entity = Span(doc, _ent['start'], _ent['end'], label=_ent['label'])
                    entity._.cui = _ent['cui']
                    entity._.detected_name = _ent['detected_name']
                    entity._.context_similarity = _ent['context_similarity']
                    entity._.id = _ent['id']
                    if 'meta_anns' in _ent:
                        entity._.meta_anns = _ent['meta_anns']
                    _ents.append(entity)
            else:
                _ents = doc.ents

            if cnf_annotation_output.get("lowercase_context", True):
                doc_tokens = [tkn.text_with_ws.lower() for tkn in list(doc)]
            else:
                doc_tokens = [tkn.text_with_ws for tkn in list(doc)]

            if cnf_annotation_output.get('doc_extended_info', False):
                # Add tokens if extended info
                out['tokens'] = doc_tokens

            context_left = cnf_annotation_output.get('context_left', -1)
            context_right = cnf_annotation_output.get('context_right', -1)
            doc_extended_info = cnf_annotation_output.get('doc_extended_info', False)

            for _, ent in enumerate(_ents):
                cui = str(ent._.cui)
                if not only_cui:
                    out_ent['pretty_name'] = self.cdb.get_name(cui)
                    out_ent['cui'] = cui
                    out_ent['type_ids'] = list(self.cdb.cui2type_ids.get(cui, ''))
                    out_ent['types'] = [self.cdb.addl_info['type_id2name'].get(tui, '') for tui in out_ent['type_ids']]
                    out_ent['source_value'] = ent.text
                    out_ent['detected_name'] = str(ent._.detected_name)
                    out_ent['acc'] = float(ent._.context_similarity)
                    out_ent['context_similarity'] = float(ent._.context_similarity)
                    out_ent['start'] = ent.start_char
                    out_ent['end'] = ent.end_char
                    for addl in addl_info:
                        tmp = self.cdb.addl_info.get(addl, {}).get(cui, [])
                        out_ent[addl.split("2")[-1]] = list(tmp) if type(tmp) == set else tmp
                    out_ent['id'] = ent._.id
                    out_ent['meta_anns'] = {}

                    if doc_extended_info:
                        out_ent['start_tkn'] = ent.start
                        out_ent['end_tkn'] = ent.end

                    if context_left > 0 and context_right > 0:
                        out_ent['context_left'] = doc_tokens[max(ent.start - context_left, 0):ent.start]
                        out_ent['context_right'] = doc_tokens[ent.end:min(ent.end + context_right, len(doc_tokens))]
                        out_ent['context_center'] = doc_tokens[ent.start:ent.end]

                    if hasattr(ent._, 'meta_anns') and ent._.meta_anns:
                        out_ent['meta_anns'] = ent._.meta_anns

                    out['entities'][out_ent['id']] = dict(out_ent)
                else:
                    out['entities'][ent._.id] = cui

            if cnf_annotation_output.get('include_text_in_output', False) or out_with_text:
                out['text'] = doc.text
        return out

    def _get_trimmed_text(self, text: Optional[str]) -> str:
        return text[0:self.config.preprocessing.get('max_document_length')] if text is not None and len(text) > 0 else ""

    def _generate_trimmed_texts(self, texts: Union[Iterable[str], Iterable[Tuple]]) -> Iterable[str]:
        text_: str
        for text in texts:
            text_ = text[1] if isinstance(text, tuple) else text
            yield self._get_trimmed_text(text_)

    def _get_trimmed_texts(self, texts: Union[Iterable[str], Iterable[Tuple]]) -> List[str]:
        trimmed: List = []
        text_: str
        for text in texts:
            text_ = text[1] if isinstance(text, tuple) else text
            trimmed.append(self._get_trimmed_text(text_))
        return trimmed

    @staticmethod
    def _pipe_error_handler(proc_name: str, proc: "Pipe", docs: List[Doc], e: Exception) -> None:
        CAT.log.warning("Exception raised when applying component %s to a batch of docs.", proc_name)
        CAT.log.warning(e, exc_info=True, stack_info=True)
        if docs is not None:
            CAT.log.warning("Docs contained in the batch:")
            for doc in docs:
                if hasattr(doc, "text"):
                    CAT.log.warning("%s...", doc.text[:50])

    @staticmethod
    def _get_doc_annotations(doc: Doc):
        if type(doc['annotations']) == list:
            return doc['annotations']
        if type(doc['annotations']) == dict:
            return doc['annotations'].values()
        return None

    def destroy_pipe(self):
        self.pipe.destroy()



import sys
sys.path.insert(0, '/home/ubuntu/projects/MedCAT/')
import os
import json
from django.shortcuts import render
from django.http import StreamingHttpResponse, HttpResponse
from wsgiref.util import FileWrapper
from medcat.cat import CAT
from medcat.cdb import CDB
from medcat.utils.helpers import doc2html
from medcat.vocab import Vocab
from urllib.request import urlretrieve, urlopen
from urllib.error import HTTPError
#from medcat.meta_cat import MetaCAT
from .models import *
from .forms import DownloaderForm

AUTH_CALLBACK_SERVICE = 'https://medcat.rosalind.kcl.ac.uk/auth-callback'
VALIDATION_BASE_URL = 'https://uts-ws.nlm.nih.gov/rest/isValidServiceValidate'
VALIDATION_LOGIN_URL = f'https://uts.nlm.nih.gov/uts/login?service={AUTH_CALLBACK_SERVICE}'

model_pack_path = os.getenv('MODEL_PACK_PATH', 'models/medmen_wstatus_2021_oct.zip')

try:
    cat = CAT.load_model_pack(model_pack_path)
except Exception as e:
    print(str(e))

def get_html_and_json(text):
    doc = cat(text)

    a = json.loads(cat.get_json(text))
    for id, ent in a['annotations'].items():
        new_ent = {}
        for key in ent.keys():
            if key == 'pretty_name':
                new_ent['Pretty Name'] = ent[key]
            if key == 'icd10':
                icd10 = ent.get('icd10', [])
                new_ent['ICD-10 Code'] = icd10[-1] if icd10 else '-'
            if key == 'cui':
                new_ent['Identifier'] = ent[key]
            if key == 'types':
                new_ent['Type'] = ", ".join(ent[key])
            if key == 'acc':
                new_ent['Confidence Score'] = ent[key]
            if key == 'start':
                new_ent['Start Index'] = ent[key]
            if key == 'end':
                new_ent['End Index'] = ent[key]
            if key == 'id':
                new_ent['id'] = ent[key]
            if key == 'meta_anns':
                meta_anns = ent.get("meta_anns", {})
                if meta_anns:
                    for meta_ann in meta_anns.keys():
                        new_ent[meta_ann] = meta_anns[meta_ann]['value']

        a['annotations'][id] = new_ent

    doc_json = json.dumps(a)
    uploaded_text = UploadedText()
    uploaded_text.text = str(text)
    uploaded_text.save()

    return doc2html(doc), doc_json


def show_annotations(request):
    context = {}
    context['doc_json'] = '{"msg": "No documents yet"}'

    if request.POST and 'text' in request.POST:
        doc_html, doc_json = get_html_and_json(request.POST['text'])

        context['doc_html'] = doc_html
        context['doc_json'] = doc_json
        context['text'] = request.POST['text']
    return render(request, 'train_annotations.html', context=context)


def validate_umls_user(request):
    ticket = request.GET.get('ticket', '')
    validate_url = f'{VALIDATION_BASE_URL}?service={AUTH_CALLBACK_SERVICE}&ticket={ticket}'
    try:
        is_valid = urlopen(validate_url, timeout=10).read().decode('utf-8')
        context = {
            'is_valid': is_valid == 'true'
        }
        if is_valid == 'true':
            context['message'] = 'License verified! Please fill in the following form before downloading models.'
            context['downloader_form'] = DownloaderForm(MedcatModel.objects.all())
        else:
            context['message'] = f'License not found. Please request or renew your UMLS Metathesaurus License. If you think you have got the license, try {VALIDATION_LOGIN_URL} again.'
    except HTTPError:
        context = {
            'is_valid': False,
            'message': 'Something went wrong. Please try again.'
        }
    finally:
        return render(request, 'umls_user_validation.html', context=context)


def download_model(request):
    if request.method == 'POST':
        downloader_form = DownloaderForm(MedcatModel.objects.all(), request.POST)
        if downloader_form.is_valid():
            downloader_form.save()
            mp_name = downloader_form.cleaned_data['modelpack']
            model = MedcatModel.objects.get(model_name=mp_name)
            if model is not None:
                mp_path = model.model_file.path
            else:
                return HttpResponse(f'Error: Unknown model "{downloader_form.modelpack}"')
            resp = StreamingHttpResponse(FileWrapper(open(mp_path, 'rb')))
            resp['Content-Type'] = 'application/zip'
            resp['Content-Length'] = os.path.getsize(mp_path)
            resp['Content-Disposition'] = f'attachment; filename={os.path.basename(mp_path)}'
            return resp
        else:
            context = {
                'is_valid': True,
                'downloader_form': downloader_form,
                'message': 'All non-optional fields must be filled out:'
            }
            return render(request, 'umls_user_validation.html', context=context)
    else:
        return HttpResponse('Erorr: Unknown HTTP method.')



from medcat.cdb_maker import CDBMaker
from medcat.config import Config, weighted_average
from functools import partial
import numpy as np
import logging
import os

config = Config()
config.general['log_level'] = logging.INFO
config.general['spacy_model'] = 'en_core_sci_lg'
maker = CDBMaker(config)

# Building a new CDB from two files (full_build)
csvs = ['./tmp_medmentions.csv']
cdb = maker.prepare_csvs(csvs, full_build=True)

cdb.save("./tmp_cdb.dat")


from medcat.vocab import Vocab
from medcat.cdb import CDB
from medcat.cat import CAT

vocab_path = "./tmp_vocab.dat"
if not os.path.exists(vocab_path):
    import requests
    tmp = requests.get("https://s3-eu-west-1.amazonaws.com/zkcl/vocab.dat")
    with open(vocab_path, 'wb') as f:
        f.write(tmp.content)

config = Config()
cdb = CDB.load("./tmp_cdb.dat", config=config)
vocab = Vocab.load(vocab_path)

cdb.reset_training()

cat = CAT(cdb=cdb, config=cdb.config, vocab=vocab)
cat.config.ner['min_name_len'] = 3
cat.config.ner['upper_case_limit_len'] = 3
cat.config.linking['disamb_length_limit'] = 3
cat.config.linking['filters'] = {'cuis': set()}
cat.config.linking['train_count_threshold'] = -1
cat.config.linking['context_vector_sizes'] = {'xlong': 27, 'long': 18, 'medium': 9, 'short': 3}
cat.config.linking['context_vector_weights'] = {'xlong': 0, 'long': 0.4, 'medium': 0.4, 'short': 0.2}
cat.config.linking['weighted_average_function'] = partial(weighted_average, factor=0.0004)
cat.config.linking['similarity_threshold_type'] = 'dynamic'
cat.config.linking['similarity_threshold'] = 0.35
cat.config.linking['calculate_dynamic_threshold'] = True

cat.train(df.text.values, fine_tune=True)


cdb.config.general['spacy_disabled_components'] = ['ner', 'parser', 'vectors', 'textcat',
                                                      'entity_linker', 'sentencizer', 'entity_ruler', 'merge_noun_chunks',
                                                                                                    'merge_entities', 'merge_subtokens']

%load_ext autoreload
%autoreload 2

# Train
_ = cat.train(open("./tmp_medmentions_text_only.txt", 'r'), fine_tune=False)

_ = cat.train_supervised("/home/ubuntu/data/medmentions/medmentions.json", reset_cui_count=True, nepochs=13, train_from_false_positives=True, print_stats=3, test_size=0.1)
cdb.save("/home/ubuntu/data/umls/2020ab/cdb_trained_medmen.dat")


_ = cat.train_supervised("/home/ubuntu/data/medmentions/medmentions.json", reset_cui_count=False, nepochs=13, train_from_false_positives=True, print_stats=3, test_size=0)

cat = CAT(cdb=cdb, config=cdb.config, vocab=vocab)
cat.config.linking['similarity_threshold'] = 0.1
cat.config.ner['min_name_len'] = 2
cat.config.ner['upper_case_limit_len'] = 1
cat.config.linking['train_count_threshold'] = -2
cat.config.linking['filters']['cuis'] = set()
cat.config.linking['context_vector_sizes'] = {'xlong': 27, 'long': 18, 'medium': 9, 'short': 3}
cat.config.linking['context_vector_weights'] = {'xlong': 0.1, 'long': 0.4, 'medium': 0.4, 'short': 0.1}
cat.config.linking['similarity_threshold_type'] = 'static'

cat.config.linking['similarity_threshold_type'] = 'dynamic'
cat.config.linking['similarity_threshold'] = 0.35
cat.config.linking['calculate_dynamic_threshold'] = True


# Print some stats
_ = cat._print_stats(data)

#Epoch: 0, Prec: 0.4331506351144245, Rec: 0.5207520064957372, F1: 0.47292889758643175
#p: 0.421  r: 0.507  f1: 0.460


# Remove all names that are numbers
for name in list(cdb.name2cuis.keys()):
    if name.replace(".", '').replace("~", '').replace(",", '').replace(":", '').replace("-", '').isnumeric():
        del cdb.name2cuis[name]
        print(name)


for name in list(cdb.name2cuis.keys()):
    if len(name) < 7 and (not name.isalpha()) and len(re.sub("[^A-Za-z]*", '', name)) < 2:
        del cdb.name2cuis[name]
        print(name)




# RUN SUPER
cdb = CDB.load("./tmp_cdb.dat")
vocab = Vocab.load(vocab_path)
cat = CAT(cdb=cdb, config=cdb.config, vocab=vocab)


# Train supervised
cdb.reset_cui_count()
cat.config.ner['uppe_case_limit_len'] = 1
cat.config.ner['min_name_len'] = 1
data_path = "./tmp_medmentions.json"
_ = cat.train_supervised(data_path, use_cui_doc_limit=True, nepochs=30, devalue_others=True, test_size=0.2)


cdb = maker.prepare_csvs(csv_paths=csvs)
cdb.save("/home/ubuntu/data/umls/2020ab/cdb_vbg.dat")



# Mapped on top-level of 2001 NHS Data Dictionary; https://datadictionary.nhs.uk/data_elements/ethnic_category.html 
ethnicity_map = {'Algerian': 'Black',
 'Any Other Group': 'Other',
 'Asian and Chinese': 'Asian',
 'Bangladeshi': 'Asian',
 'Black African': 'Black',
 'Black British': 'Black',
 'British': 'White',
 'Caribbean': 'Black',
 'Chinese': 'Asian',
 'Cypriot (Part nt st)': 'White',
 'Ecuadorian': 'Other',
 'English': 'White',
 'Ethiopian': 'Black',
 'Filipino': 'Asian',
 'Ghanaian': 'Black',
 'Greek Cypriot': 'White',
 'Indian/British India': 'Asian',
 'Iranian': 'Other',
 'Italian': 'White',
 'Mixed Black': 'Black',
 'Mixed Caribbean': 'Black',
 'Nigerian': 'Black',
 'Not Given': 'Unknown',
 'Not Specified': 'Unknown',
 'Not Stated': 'Unknown',
 'OTHER ASIAN BACKGROU': 'Asian',
 'Other Asian Unspecif': 'Asian',
 'OTHER BLACK BACKGROU': 'Black',
 'Other Black Unspecif': 'Black',
 'Other Ethnic Group': 'Other',
 'Other Latin American': 'Other',
 'OTHER WHITE BACK GRO': 'White',
 'Other White Unspecif': 'White',
 'Other White/Mixed Eu': 'White',
 'Pakistani/British Pa': 'Asian',
 'Portuguese': 'White',
 'Somali': 'Black',
 'Spanish': 'White',
 'Sri Lankan': 'Asian',
 'Sudanese': 'Black',
 'Turkish': 'Other',
 'Ugandan': 'Black',
 'Vietnamese': 'Asian',
 'White Irish': 'White',
 'Former USSR Rep': 'White',
 'POLISH': 'White',
 'Iraqi': 'Other',
 'Albanian': 'Other',
 'Columbian': 'Other',
 'Scottish': 'White',
 'Not stated': 'Unknown',
 'OTHER MIXED BACKGROU': 'Mixed',
 'Welsh': 'White',
 'British Asian': 'Asian',
 'Caribbean Asian': 'Asian',
 'Eritrean': 'Black',
 'Turkish Cypriot': 'Other',
 'Sinhalese': 'Asian',
 'White and Asian': 'Asian',
 'Other Mixed': 'Mixed',
 'Mixed Asian': 'Asian',
 'Greek': 'White',
 'Arab': 'Other',
 'MULTIPLE CODES': 'MULTIPLE CODES',
 'Irish': 'White',
 'Japanese': 'Asian',
 'Middle East': 'Other',
 'Croatian': 'White',
 'Black and Asian': 'Mixed',
 'Black and White': 'Mixed'}

# Mapped on bottom-level of 2001 NHS Data Dictionary; https://datadictionary.nhs.uk/data_elements/ethnic_category.html 
ethnicity_map_detail = {'Algerian': 'Black or Black British - African',
 'Any Other Group': 'Other Ethnic Groups - Any other ethnic group',
 'Asian and Chinese': 'Other Ethnic Groups - Chinese',
 'Bangladeshi': 'Asian or Asian British - Pakistani',
 'Black African': 'Black or Black British - African',
 'Black British': 'Black or Black British - Any Other Black background',
 'British': 'White - British',
 'Caribbean': 'Black or Black British - Caribbean',
 'Chinese': 'Other Ethnic Groups - Chinese',
 'Cypriot (Part nt st)': 'White - Any other White background',
 'Ecuadorian': 'Other Ethnic Groups - Any other ethnic group',
 'English': 'White - British',
 'Ethiopian': 'Black or Black British - African',
 'Filipino': 'Asian or Asian British - Any other Asian background',
 'Ghanaian': 'Black or Black British - African',
 'Greek Cypriot': 'White - Any other White background',
 'Indian/British India': 'Asian or Asian British - Indian',
 'Iranian': 'Other Ethnic Groups - Any other ethnic group',
 'Italian': 'White - Any other White background',
 'Mixed Black': 'Black or Black British - Any other Black background',
 'Mixed Caribbean': 'Black or Black British - Caribbean',
 'Nigerian': 'Black or Black British - African',
 'Not Given': 'Not stated',
 'Not Specified': 'Not stated',
 'Not Stated': 'Not stated',
 'OTHER ASIAN BACKGROU': 'Asian or Asian British - Any other Asian background',
 'Other Asian Unspecif': 'Asian or Asian British - Any other Asian background',
 'OTHER BLACK BACKGROU': 'Black or Black British - Any Other Black background',
 'Other Black Unspecif': 'Black or Black British - Any Other Black background',
 'Other Ethnic Group': 'Other Ethnic Groups - Any other ethnic group',
 'Other Latin American': 'Other Ethnic Groups - Any other ethnic group',
 'OTHER WHITE BACK GRO': 'White - Any other White background',
 'Other White Unspecif': 'White - Any other White background',
 'Other White/Mixed Eu': 'White - Any other White background',
 'Pakistani/British Pa': 'Asian or Asian British - Pakistani',
 'Portuguese': 'White - Any other White background',
 'Somali': 'Black or Black British - African',
 'Spanish': 'White - Any other White background',
 'Sri Lankan': 'Asian or Asian British - Any other Asian background',
 'Sudanese': 'Black or Black British - African',
 'Turkish': 'Other Ethnic Groups - Any other ethnic group',
 'Ugandan': 'Black or Black British - African',
 'Vietnamese': 'Other Ethnic Groups - Any other ethnic group',
 'White Irish': 'White - Irish',
 'Former USSR Rep': 'White - Any other White background',
 'POLISH': 'White - Any other White background',
 'Iraqi': 'Other Ethnic Groups - Any other ethnic group',
 'Albanian': 'White - Any other White background',
 'Columbian': 'Other Ethnic Groups - Any other ethnic group',
 'Scottish': 'White - British',
 'Not stated': 'Not stated',
 'OTHER MIXED BACKGROU': 'Mixed - Any other mixed background',
 'Welsh': 'White - British',
 'British Asian': 'Asian or Asian British - Any other Asian background',
 'Caribbean Asian': 'Mixed - Any other mixed background',
 'Eritrean': 'Black or Black British - African',
 'Turkish Cypriot': 'Other Ethnic Groups - Any other ethnic group',
 'Sinhalese': 'Asian or Asian British - Any other Asian background',
 'White and Asian': 'Mixed - White and Asian',
 'Other Mixed': 'Mixed - Any other mixed background',
 'Mixed Asian': 'Mixed - Any other mixed background',
 'Greek': 'White - Any other White background',
 'Arab': 'Other Ethnic Groups - Any other ethnic group',
 'MULTIPLE CODES': 'MULTIPLE CODES',
 'Irish': 'White - Irish',
 'Japanese': 'Other Ethnic Groups - Any other ethnic group',
 'Middle East': 'Other Ethnic Groups - Any other ethnic group',
 'Croatian': 'White - Any other White background',
 'Black and Asian': 'Mixed - White and Asian',
 'Black and White': 'Mixed - Any other mixed background'}



from django.contrib import admin
from django.urls import path
from .views import *

urlpatterns = [
    path('', show_annotations, name='train_annotations'),
    path('auth-callback', validate_umls_user, name='validate-umls-user'),
    path('download-model', download_model, name="download-model")
]



import dill
from typing import Optional, Dict
from transformers.tokenization_utils_base import PreTrainedTokenizerBase


class TokenizerNER(object):
    r'''

    Args:
        hf_tokenizer
            Must be able to return token offsets
        max_len:
            Max sequence length, if longer it will be split into multiple examples
    '''

    def __init__(self,
                 hf_tokenizer: Optional[PreTrainedTokenizerBase] = None,
                 max_len: int = 512,
                 id2type: Optional[Dict] = None,
                 cui2name: Optional[Dict] = None) -> None:
        self.hf_tokenizer = hf_tokenizer
        self.max_len = max_len
        self.label_map = {'O': 0, 'X': 1}
        self.id2type = id2type
        self.cui2name = cui2name

    def calculate_label_map(self, dataset: Dict) -> None:
        for cuis in dataset['ent_cuis']:
            for cui in cuis:
                if cui not in self.label_map:
                    self.label_map[cui] = len(self.label_map)

    def encode(self, examples: Dict, ignore_subwords: bool = False) -> Dict:
        r''' Used with huggingface datasets map function to convert medcat_ner dataset into the
        appropriate form for NER with BERT. It will split long text segments into max_len sequences.

        Args:
            examples:
                Stream of examples
            ignore_subwords:
                If set to `True` subwords of any token will get the special label `X`
        '''
        self.hf_tokenizer = self.ensure_tokenizer()
        old_ids = examples['id']
        examples['input_ids'] = []
        examples['labels'] = []
        examples['id'] = []

        for _ind, example in enumerate(zip(examples['text'], examples['ent_starts'],
                examples['ent_ends'], examples['ent_cuis'])):
            tokens = self.hf_tokenizer(example[0], return_offsets_mapping=True, add_special_tokens=False)
            entities = [(start, end, cui) for start, end, cui in zip(example[1],
                        example[2], example[3])]
            entities.sort(key=lambda x: x[0])
            input_ids = []
            labels = []

            tkn_part_of_entity = False
            for ind in range(len(tokens['offset_mapping'])):
                offset = tokens['offset_mapping'][ind]
                input_ids.append(tokens['input_ids'][ind])

                if entities and (offset[0] >= entities[0][0] and offset[1] <= entities[0][1]):
                    # Means this token is part of entity at position 0
                    tkn_part_of_entity = True
                    if not ignore_subwords or (self.id2type is not None and self.id2type[tokens['input_ids'][ind]] == 'start'):
                        labels.append(self.label_map[entities[0][2]])
                    else:
                        labels.append(self.label_map['X'])

                    if entities[0][1] <= offset[1]:
                        # If it is the last token of the entity, remove the entity as it is done
                        del entities[0]
                        tkn_part_of_entity = False # Set this so the next token is not removed

                else:
                    if tkn_part_of_entity:
                        del entities[0]
                        tkn_part_of_entity = False

                    if not ignore_subwords or (self.id2type is not None and self.id2type[tokens['input_ids'][ind]] == 'start'):
                        labels.append(self.label_map["O"])
                    else:
                        labels.append(self.label_map['X'])

                if len(input_ids) >= self.max_len:
                    # Split into multiple examples if too long
                    examples['input_ids'].append(input_ids)
                    examples['labels'].append(labels)
                    examples['id'].append(old_ids[_ind])

                    input_ids = []
                    labels = []

            if input_ids:
                examples['input_ids'].append(input_ids)
                examples['labels'].append(labels)
                examples['id'].append(old_ids[_ind])

        return examples

    def save(self, path: str) -> None:
        with open(path, 'wb') as f:
            dill.dump(self.__dict__, f)

    def ensure_tokenizer(self) -> PreTrainedTokenizerBase:
        if self.hf_tokenizer is None:
            raise ValueError("The tokenizer is not loaded yet")
        return self.hf_tokenizer

    @classmethod
    def load(cls, path: str) -> 'TokenizerNER':
        tokenizer = cls()
        with open(path, 'rb') as f:
            d = dill.load(f)
            for k in tokenizer.__dict__:
                if k in d:
                    tokenizer.__dict__[k] = d[k]
        return tokenizer



import os
from pkg_resources import get_distribution, DistributionNotFound

name = 'medcat'

try:
    distribution = get_distribution(name)
    distribution_location = os.path.normcase(distribution.location)
    init_location = os.path.normcase(__file__)
    if not init_location.startswith(os.path.join(distribution_location, name)):
        raise DistributionNotFound
except DistributionNotFound:
    __version__ = f'Please install the package: {name}'
else:
    __version__ = distribution.version



import random
import logging

from enum import Enum, auto
from spacy.tokens import Span, Doc
from typing import Optional, List, Dict
from medcat.utils.filters import check_filters
from medcat.linking.vector_context_model import ContextModel
from medcat.pipeline.pipe_runner import PipeRunner
from medcat.cdb import CDB
from medcat.vocab import Vocab
from medcat.config import Config


class LabelStyle(Enum):
    short = auto()
    long = auto()


class Linker(PipeRunner):
    r''' Link to a biomedical database.

    Args:
        cdb
        vocab
        config
    '''
    log = logging.getLogger(__name__)

    # Custom pipeline component name
    name = 'cat_linker'

    # Override
    def __init__(self, cdb: CDB, vocab: Vocab, config: Config) -> None:
        self.cdb = cdb
        self.vocab = vocab
        self.config = config
        self.context_model = ContextModel(self.cdb, self.vocab, self.config)
        # Counter for how often did a pair (name,cui) appear and was used during training
        self.train_counter: Dict = {}
        super().__init__(self.config.general['workers'])

    def _train(self, cui: str, entity: Span, doc: Doc, add_negative: bool = True) -> None:
        name = "{} - {}".format(entity._.detected_name, cui)
        """ TODO: Disable for now
        if self.train_counter.get(name, 0) > self.config.linking['subsample_after']:
            if random.random() < 1 / math.sqrt(self.train_counter.get(name) - self.config.linking['subsample_after']):
                self.context_model.train(cui, entity, doc, negative=False)
                if add_negative and self.config.linking['negative_probability'] >= random.random():
                    self.context_model.train_using_negative_sampling(cui)
                self.train_counter[name] = self.train_counter.get(name, 0) + 1
        else:
        """
        # Always train
        self.context_model.train(cui, entity, doc, negative=False)
        if add_negative and self.config.linking['negative_probability'] >= random.random():
            self.context_model.train_using_negative_sampling(cui)
        self.train_counter[name] = self.train_counter.get(name, 0) + 1

    # Override
    def __call__(self, doc: Doc) -> Doc:
        r'''
        '''
        cnf_l = self.config.linking
        linked_entities = []

        if cnf_l["train"]:
            # Run training
            for entity in doc._.ents:
                # Check does it have a detected name
                if entity._.detected_name is not None:
                    name = entity._.detected_name
                    cuis = entity._.link_candidates

                    if len(name) >= cnf_l['disamb_length_limit']:
                        if len(cuis) == 1:
                            # N - means name must be disambiguated, is not the prefered
                            #name of the concept, links to other concepts also.
                            if self.cdb.name2cuis2status[name][cuis[0]] != 'N':
                                self._train(cui=cuis[0], entity=entity, doc=doc)
                                entity._.cui = cuis[0]
                                entity._.context_similarity = 1
                                linked_entities.append(entity)
                        else:
                            for cui in cuis:
                                if self.cdb.name2cuis2status[name][cui] in {'P', 'PD'}:
                                    self._train(cui=cui, entity=entity, doc=doc)
                                    # It should not be possible that one name is 'P' for two CUIs,
                                    #but it can happen - and we do not care.
                                    entity._.cui = cui
                                    entity._.context_similarity = 1
                                    linked_entities.append(entity)
        else:
            for entity in doc._.ents:
                self.log.debug("Linker started with entity: %s", entity)
                # Check does it have a detected name
                if entity._.link_candidates is not None:
                    if entity._.detected_name is not None:
                        name = entity._.detected_name
                        cuis = entity._.link_candidates

                        if len(cuis) > 0:
                            do_disambiguate = False
                            if len(name) < cnf_l['disamb_length_limit']:
                                do_disambiguate = True
                            elif len(cuis) == 1 and self.cdb.name2cuis2status[name][cuis[0]] in {'N', 'PD'}:
                                # PD means it is preferred but should still be disambiguated and N is disamb always
                                do_disambiguate = True
                            elif len(cuis) > 1:
                                do_disambiguate = True

                            if do_disambiguate:
                                cui, context_similarity = self.context_model.disambiguate(cuis, entity, name, doc)
                            else:
                                cui = cuis[0]
                                if self.config.linking['always_calculate_similarity']:
                                    context_similarity = self.context_model.similarity(cui, entity, doc)
                                else:
                                    context_similarity = 1 # Direct link, no care for similarity
                    else:
                        # No name detected, just disambiguate
                        cui, context_similarity = self.context_model.disambiguate(entity._.link_candidates, entity, 'unk-unk', doc)

                    # Add the annotation if it exists and if above threshold and in filters
                    if cui and check_filters(cui, self.config.linking['filters']):
                        th_type = self.config.linking.get('similarity_threshold_type', 'static')
                        if (th_type == 'static' and context_similarity >= self.config.linking['similarity_threshold']) or \
                           (th_type == 'dynamic' and context_similarity >= self.cdb.cui2average_confidence[cui] * self.config.linking['similarity_threshold']):
                            entity._.cui = cui
                            entity._.context_similarity = context_similarity
                            linked_entities.append(entity)

        doc._.ents = linked_entities
        self._create_main_ann(doc)

        if self.config.general['make_pretty_labels'] is not None:
            self._make_pretty_labels(doc, LabelStyle[self.config.general['make_pretty_labels']])

        if self.config.general['map_cui_to_group'] is not None and self.cdb.addl_info.get('cui2group', {}):
            self._map_ents_to_groups(doc)

        return doc

    def _map_ents_to_groups(self, doc: Doc) -> None:
        for ent in doc.ents:
            ent._.cui = self.cdb.addl_info['cui2group'].get(ent._.cui, ent._.cui)

    def _make_pretty_labels(self, doc: Doc, style: Optional[LabelStyle] = None) -> None:
        ents = list(doc.ents)

        n_ents = []
        for ent in ents:
            if style == LabelStyle.short:
                label = ent._.cui
            elif style == LabelStyle.long:
                label = "{} | {} | {:.2f}".format(ent._.cui, self.cdb.get_name(ent._.cui), ent._.context_similarity)
            else:
                label = 'concept'

            n_ent = Span(doc, ent.start, ent.end, label)
            for attr in ent._.__dict__['_extensions'].keys():
                setattr(n_ent._, attr, getattr(ent._, attr))
            n_ents.append(n_ent)

        doc.ents = n_ents

    def _create_main_ann(self, doc: Doc, tuis: Optional[List] = None) -> None:
        # TODO: Separate into another piece of the pipeline
        """ Creates annotation in the spacy ents list
        from all the annotations for this document.

        doc:  spacy document
        """
        doc._.ents.sort(key=lambda x: len(x.text), reverse=True)

        tkns_in = set()
        main_anns = []
        for ent in doc._.ents:
            if tuis is None or ent._.tui in tuis:
                to_add = True
                for tkn in ent:
                    if tkn in tkns_in:
                        to_add = False
                if to_add:
                    for tkn in ent:
                        tkns_in.add(tkn)
                    main_anns.append(ent)

        doc.ents = list(doc.ents) + main_anns



from gensim.matutils import unitvec as g_unitvec
import numpy as np


def unitvec(arr):
    return g_unitvec(np.array(arr))


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def intersect_nonempty_set(base, target):
    if isinstance(target, set) and target:
        result = base.intersection(target)
    else:
        result = base

    # If result is empty set to empty
    if not result:
        result = {'empty'}

    return result



from typing import Any, Dict, List
import torch


class CollateAndPadNER(object):
    def __init__(self, pad_id):
        self.pad_id = pad_id

    def __call__(self, features: List[Any]) -> Dict[str, torch.Tensor]:
        batch = {}

        max_len = max([len(f['input_ids']) for f in features])
        batch['input_ids'] = torch.tensor([f['input_ids'][0:max_len] + [self.pad_id] *
            max(0, max_len - len(f['input_ids'])) for f in features], dtype=torch.long)
        batch['labels'] = torch.tensor([f['labels'][0:max_len] + [-100] *
            max(0, max_len - len(f['labels'])) for f in features], dtype=torch.long)
        batch['attention_mask'] = batch['input_ids'] != self.pad_id

        return batch



import json
import os
import sys
import unittest
import tempfile
from medcat.vocab import Vocab
from medcat.cdb import CDB
from medcat.cat import CAT
from medcat.utils.checkpoint import Checkpoint


class CATTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.cdb = CDB.load(os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "examples", "cdb.dat"))
        cls.vocab = Vocab.load(os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "examples", "vocab.dat"))
        cls.cdb.config.general["spacy_model"] = "en_core_web_md"
        cls.cdb.config.ner['min_name_len'] = 2
        cls.cdb.config.ner['upper_case_limit_len'] = 3
        cls.cdb.config.general['spell_check'] = True
        cls.cdb.config.linking['train_count_threshold'] = 10
        cls.cdb.config.linking['similarity_threshold'] = 0.3
        cls.cdb.config.linking['train'] = True
        cls.cdb.config.linking['disamb_length_limit'] = 5
        cls.cdb.config.general['full_unlink'] = True
        cls.undertest = CAT(cdb=cls.cdb, config=cls.cdb.config, vocab=cls.vocab)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.undertest.destroy_pipe()

    def tearDown(self) -> None:
        self.cdb.config.annotation_output['include_text_in_output'] = False

    def test_callable_with_single_text(self):
        text = "The dog is sitting outside the house."
        doc = self.undertest(text)
        self.assertEqual(text, doc.text)

    def test_callable_with_single_empty_text(self):
        self.assertIsNone(self.undertest(""))

    def test_callable_with_single_none_text(self):
        self.assertIsNone(self.undertest(None))

    def test_multiprocessing(self):
        in_data = [
            (1, "The dog is sitting outside the house and second csv."),
            (2, ""),
            (3, None)
        ]
        out = self.undertest.multiprocessing(in_data, nproc=1)

        self.assertEqual(3, len(out))
        self.assertEqual(1, len(out[1]['entities']))
        self.assertEqual(0, len(out[2]['entities']))
        self.assertEqual(0, len(out[3]['entities']))

    def test_multiprocessing_pipe(self):
        in_data = [
            (1, "The dog is sitting outside the house and second csv."),
            (2, "The dog is sitting outside the house."),
            (3, "The dog is sitting outside the house."),
        ]
        out = self.undertest.multiprocessing_pipe(in_data, nproc=2, return_dict=False)
        self.assertTrue(type(out) == list)
        self.assertEqual(3, len(out))
        self.assertEqual(1, out[0][0])
        self.assertEqual('second csv', out[0][1]['entities'][0]['source_value'])
        self.assertEqual(2, out[1][0])
        self.assertEqual({'entities': {}, 'tokens': []}, out[1][1])
        self.assertEqual(3, out[2][0])
        self.assertEqual({'entities': {}, 'tokens': []}, out[2][1])

    def test_multiprocessing_pipe_with_malformed_texts(self):
        in_data = [
            (1, "The dog is sitting outside the house."),
            (2, ""),
            (3, None),
        ]
        out = self.undertest.multiprocessing_pipe(in_data, nproc=1, batch_size=1, return_dict=False)
        self.assertTrue(type(out) == list)
        self.assertEqual(3, len(out))
        self.assertEqual(1, out[0][0])
        self.assertEqual({'entities': {}, 'tokens': []}, out[0][1])
        self.assertEqual(2, out[1][0])
        self.assertEqual({'entities': {}, 'tokens': []}, out[1][1])
        self.assertEqual(3, out[2][0])
        self.assertEqual({'entities': {}, 'tokens': []}, out[2][1])

    def test_multiprocessing_pipe_return_dict(self):
        in_data = [
            (1, "The dog is sitting outside the house."),
            (2, "The dog is sitting outside the house."),
            (3, "The dog is sitting outside the house.")
        ]
        out = self.undertest.multiprocessing_pipe(in_data, nproc=2, return_dict=True)
        self.assertTrue(type(out) == dict)
        self.assertEqual(3, len(out))
        self.assertEqual({'entities': {}, 'tokens': []}, out[1])
        self.assertEqual({'entities': {}, 'tokens': []}, out[2])
        self.assertEqual({'entities': {}, 'tokens': []}, out[3])

    def test_train(self):
        ckpt_steps = 2
        nepochs = 3
        ckpt_dir_path = tempfile.TemporaryDirectory().name
        checkpoint = Checkpoint(dir_path=ckpt_dir_path, steps=ckpt_steps)
        self.undertest.cdb.print_stats()
        self.undertest.train(["The dog is not a house"] * 20, nepochs=nepochs, checkpoint=checkpoint)
        self.undertest.cdb.print_stats()
        checkpoints = [f for f in os.listdir(ckpt_dir_path) if "checkpoint-" in f]

        self.assertEqual(1, len(checkpoints))
        self.assertEqual(f"checkpoint-{ckpt_steps}-{nepochs * 20}", checkpoints[0])

    def test_resume_training(self):
        nepochs_train = 1
        nepochs_retrain = 1
        ckpt_steps = 3
        ckpt_dir_path = tempfile.TemporaryDirectory().name
        checkpoint = Checkpoint(dir_path=ckpt_dir_path, steps=ckpt_steps, max_to_keep=sys.maxsize)
        self.undertest.cdb.print_stats()
        self.undertest.train(["The dog is not a house"] * 20,
                             nepochs=nepochs_train,
                             checkpoint=checkpoint,
                             is_resumed=False)
        self.undertest.cdb.print_stats()
        self.undertest.train(["The dog is not a house"] * 20,
                             nepochs=nepochs_train+nepochs_retrain,
                             checkpoint=checkpoint,
                             is_resumed=True)
        checkpoints = [f for f in os.listdir(ckpt_dir_path) if "checkpoint-" in f]
        self.assertEqual(13, len(checkpoints))
        self.assertTrue("checkpoint-%s-3" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-6" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-9" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-12" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-15" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-18" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-21" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-24" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-27" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-30" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-33" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-36" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-39" % ckpt_steps in checkpoints)

    def test_resume_training_on_absent_checkpoints(self):
        ckpt_dir_path = tempfile.TemporaryDirectory().name
        checkpoint = Checkpoint(dir_path=ckpt_dir_path)
        with self.assertRaises(Exception) as e:
            self.undertest.train(["The dog is not a house"] * 40, checkpoint=checkpoint, is_resumed=True)
        self.assertEqual("Checkpoints not found. You need to train from scratch.", str(e.exception))

    def test_train_keep_n_checkpoints(self):
        ckpt_steps = 2
        ckpt_dir_path = tempfile.TemporaryDirectory().name
        checkpoint = Checkpoint(dir_path=ckpt_dir_path, steps=ckpt_steps, max_to_keep=2)
        self.undertest.cdb.print_stats()
        self.undertest.train(["The dog is not a house"] * 20, checkpoint=checkpoint)
        self.undertest.cdb.print_stats()
        checkpoints = [f for f in os.listdir(ckpt_dir_path) if "checkpoint-" in f]
        self.assertEqual(2, len(checkpoints))
        self.assertTrue("checkpoint-%s-18" % ckpt_steps in checkpoints)
        self.assertTrue("checkpoint-%s-20" % ckpt_steps in checkpoints)

    def test_get_entities(self):
        text = "The dog is sitting outside the house."
        out = self.undertest.get_entities(text)
        self.assertEqual({}, out["entities"])
        self.assertEqual([], out["tokens"])
        self.assertFalse("text" in out)

    def test_get_entities_including_text(self):
        self.cdb.config.annotation_output['include_text_in_output'] = True
        text = "The dog is sitting outside the house."
        out = self.undertest.get_entities(text)
        self.assertEqual({}, out["entities"])
        self.assertEqual([], out["tokens"])
        self.assertTrue(text in out["text"])

    def test_get_entities_multi_texts(self):
        in_data = [(1, "The dog is sitting outside the house."), (2, ""), (3, "The dog is sitting outside the house.")]
        out = self.undertest.get_entities_multi_texts(in_data, n_process=2)
        self.assertEqual(3, len(out))
        self.assertFalse("text" in out[0])
        self.assertFalse("text" in out[1])
        self.assertFalse("text" in out[2])

    def test_get_entities_multi_texts_including_text(self):
        self.cdb.config.annotation_output['include_text_in_output'] = True
        in_data = [(1, "The dog is sitting outside the house."), (2, ""), (3, None)]
        out = self.undertest.get_entities_multi_texts(in_data, n_process=2)
        self.assertEqual(3, len(out))
        self.assertTrue("text" in out[0])
        self.assertFalse("text" in out[1])
        self.assertFalse("text" in out[2])

    def test_train_supervised(self):
        nepochs = 3
        num_of_documents = 27
        data_path = os.path.join(os.path.dirname(__file__), "resources", "medcat_trainer_export.json")
        ckpt_dir_path = tempfile.TemporaryDirectory().name
        checkpoint = Checkpoint(dir_path=ckpt_dir_path, steps=1, max_to_keep=sys.maxsize)
        fp, fn, tp, p, r, f1, cui_counts, examples = self.undertest.train_supervised(data_path,
                                                                                     checkpoint=checkpoint,
                                                                                     nepochs=nepochs)
        checkpoints = [f for f in os.listdir(ckpt_dir_path) if "checkpoint-" in f]
        self.assertEqual({}, fp)
        self.assertEqual({}, fn)
        self.assertEqual({}, tp)
        self.assertEqual({}, p)
        self.assertEqual({}, r)
        self.assertEqual({}, f1)
        self.assertEqual({}, cui_counts)
        self.assertEqual({}, examples)
        self.assertEqual(nepochs * num_of_documents, len(checkpoints))
        for step in range(1, nepochs * num_of_documents + 1):
            self.assertTrue(f"checkpoint-1-{step}" in checkpoints)

    def test_resume_supervised_training(self):
        nepochs_train = 1
        nepochs_retrain = 2
        num_of_documents = 27
        data_path = os.path.join(os.path.dirname(__file__), "resources", "medcat_trainer_export.json")
        ckpt_dir_path = tempfile.TemporaryDirectory().name
        checkpoint = Checkpoint(dir_path=ckpt_dir_path, steps=1, max_to_keep=sys.maxsize)
        self.undertest.train_supervised(data_path,
                                        checkpoint=checkpoint,
                                        nepochs=nepochs_train)
        fp, fn, tp, p, r, f1, cui_counts, examples = self.undertest.train_supervised(data_path,
                                                                                     checkpoint=checkpoint,
                                                                                     nepochs=nepochs_train+nepochs_retrain,
                                                                                     is_resumed=True)
        checkpoints = [f for f in os.listdir(ckpt_dir_path) if "checkpoint-" in f]
        self.assertEqual({}, fp)
        self.assertEqual({}, fn)
        self.assertEqual({}, tp)
        self.assertEqual({}, p)
        self.assertEqual({}, r)
        self.assertEqual({}, f1)
        self.assertEqual({}, cui_counts)
        self.assertEqual({}, examples)
        self.assertEqual((nepochs_train + nepochs_retrain) * num_of_documents, len(checkpoints))
        for step in range(1, (nepochs_train + nepochs_retrain) * num_of_documents):
            self.assertTrue(f"checkpoint-1-{step}" in checkpoints)

    def test_no_error_handling_on_none_input(self):
        out = self.undertest.get_entities(None)
        self.assertEqual({}, out["entities"])
        self.assertEqual([], out["tokens"])

    def test_no_error_handling_on_empty_string_input(self):
        out = self.undertest.get_entities("")
        self.assertEqual({}, out["entities"])
        self.assertEqual([], out["tokens"])

    def test_no_raise_on_single_process_with_none(self):
        out = self.undertest.get_entities_multi_texts(["The dog is sitting outside the house.", None, "The dog is sitting outside the house."], n_process=1, batch_size=2)
        self.assertEqual(3, len(out))
        self.assertEqual({}, out[0]["entities"])
        self.assertEqual([], out[0]["tokens"])
        self.assertEqual({}, out[1]["entities"])
        self.assertEqual([], out[1]["tokens"])
        self.assertEqual({}, out[2]["entities"])
        self.assertEqual([], out[2]["tokens"])

    def test_no_raise_on_single_process_with_empty_string(self):
        out = self.undertest.get_entities_multi_texts(["The dog is sitting outside the house.", "", "The dog is sitting outside the house."], n_process=1, batch_size=2)
        self.assertEqual(3, len(out))
        self.assertEqual({}, out[0]["entities"])
        self.assertEqual([], out[0]["tokens"])
        self.assertEqual({}, out[1]["entities"])
        self.assertEqual([], out[1]["tokens"])
        self.assertEqual({}, out[2]["entities"])
        self.assertEqual([], out[2]["tokens"])

    def test_error_handling_multi_processes(self):
        self.cdb.config.annotation_output['include_text_in_output'] = True
        out = self.undertest.get_entities_multi_texts([
                                           (1, "The dog is sitting outside the house 1."),
                                           (2, "The dog is sitting outside the house 2."),
                                           (3, "The dog is sitting outside the house 3."),
                                           (4, None),
                                           (5, None)], n_process=2, batch_size=2)
        self.assertEqual(5, len(out))
        self.assertEqual({}, out[0]["entities"])
        self.assertEqual([], out[0]["tokens"])
        self.assertTrue("The dog is sitting outside the house 1.", out[0]["text"])
        self.assertEqual({}, out[1]["entities"])
        self.assertEqual([], out[1]["tokens"])
        self.assertTrue("The dog is sitting outside the house 2.", out[1]["text"])
        self.assertEqual({}, out[2]["entities"])
        self.assertEqual([], out[2]["tokens"])
        self.assertTrue("The dog is sitting outside the house 3.", out[2]["text"])
        self.assertEqual({}, out[3]["entities"])
        self.assertEqual([], out[3]["tokens"])
        self.assertFalse("text" in out[3])
        self.assertEqual({}, out[4]["entities"])
        self.assertEqual([], out[4]["tokens"])
        self.assertFalse("text" in out[4])

    def test_create_model_pack(self):
        save_dir_path = tempfile.TemporaryDirectory()
        full_model_pack_name = self.undertest.create_model_pack(save_dir_path.name, model_pack_name="mp_name")
        pack = [f for f in os.listdir(save_dir_path.name)]
        self.assertTrue(full_model_pack_name in pack)
        self.assertTrue(f'{full_model_pack_name}.zip' in pack)
        contents = [f for f in os.listdir(os.path.join(save_dir_path.name, full_model_pack_name))]
        self.assertTrue("cdb.dat" in contents)
        self.assertTrue("vocab.dat" in contents)
        self.assertTrue("model_card.json" in contents)
        with open(os.path.join(save_dir_path.name, full_model_pack_name, "model_card.json")) as file:
            model_card = json.load(file)
        self.assertTrue("MedCAT Version" in model_card)

    def test_load_model_pack(self):
        save_dir_path = tempfile.TemporaryDirectory()
        full_model_pack_name = self.undertest.create_model_pack(save_dir_path.name, model_pack_name="mp_name")
        cat = self.undertest.load_model_pack(os.path.join(save_dir_path.name, f"{full_model_pack_name}.zip"))
        self.assertTrue(isinstance(cat, CAT))
        self.assertIsNotNone(cat.config.version['medcat_version'])

    def test_hashing(self):
        save_dir_path = tempfile.TemporaryDirectory()
        full_model_pack_name = self.undertest.create_model_pack(save_dir_path.name, model_pack_name="mp_name")
        cat = self.undertest.load_model_pack(os.path.join(save_dir_path.name, f"{full_model_pack_name}.zip"))
        self.assertEqual(cat.get_hash(), cat.config.version['id'])


if __name__ == '__main__':
    unittest.main()



import argparse
import logging
import yaml

from medcat.cdb_maker import CDBMaker
from medcat.utils.make_vocab import MakeVocab
from medcat.cat import CAT
from medcat.config import Config
from medcat.utils.loggers import add_handlers
from pathlib import Path

# Create Logger
logger = logging.getLogger('__package__')
logger = add_handlers(logger)
logger.setLevel(logging.INFO)


def create_cdb(concept_csv_file, medcat_config):
    """Create concept database from csv.

    Args:
        concept_csv_file (pathlib.Path):
            Path to CSV file containing all concepts and synonyms.
        medcat_config (medcat.config.Config):
            MedCAT configuration file.
    Returns:
        cdb (medcat.cdb.CDB):
            MedCAT concept database containing list of entities and synonyms, without context embeddings.
    """
    logger.info('Creating concept database from concept table')
    cdb_maker = CDBMaker(config=medcat_config)
    cdb = cdb_maker.prepare_csvs([str(concept_csv_file)], full_build=True)
    return cdb


def create_vocab(cdb, training_data_list, medcat_config, output_dir, unigram_table_size):
    """Create vocabulary for word embeddings and spell check from list of training documents and CDB.

    Args:
        cdb (medcat.cdb.CDB):
            MedCAT concept database containing list of entities and synonyms.
        training_data_list (list):
            List of example documents.
        medcat_config (medcat.config.Config):
            MedCAT configuration file.
        output_dir (pathlib.Path):
            Output directory to write vocabulary and data.txt (required to create vocabulary) to.
        unigram_table_size (int):
            Size of unigram table to be initialized before creating vocabulary.

    Returns:
        vocab (medcat.vocab.Vocab):
            MedCAT vocabulary created from CDB and training documents.
    """
    logger.info('Creating and saving vocabulary')
    make_vocab = MakeVocab(cdb=cdb, config=medcat_config)
    make_vocab.make(training_data_list, out_folder=str(output_dir))
    make_vocab.add_vectors(in_path=str(output_dir/'data.txt'), unigram_table_size=unigram_table_size)
    vocab = make_vocab.vocab
    return vocab


def train_unsupervised(cdb, vocab, medcat_config, output_dir, training_data_list):
    """Perform unsupervised training and save updated CDB.

    Although not returned explicitly in this function, the CDB will be updated with context embeddings.

    Args:
        cdb (medcat.cdb.CDB):
            MedCAT concept database containing list of entities and synonyms.
        vocab (medcat.vocab.Vocab):
            MedCAT vocabulary created from CDB and training documents.
        medcat_config (medcat.config.Config):
            MedCAT configuration file.
        output_dir (pathlib.Path):
            Output directory to write updated CDB to.
        training_data_list (list):
            List of example documents.

    Returns:
        cdb (medcat.cdb.CDB):
            MedCAT concept database containing list of entities and synonyms, as well as context embeddings.
    """
    # Create MedCAT pipeline
    cat = CAT(cdb=cdb, vocab=vocab, config=medcat_config)

    # Perform unsupervised training and add model to concept database
    logger.info('Performing unsupervised training')
    cat.train(training_data_list)

    # Save output
    logger.info('Saving updated concept database')
    cdb.save(str(output_dir / 'cdb.dat'))

    return cdb


def create_models(config_file):
    """Create MedCAT CDB and Vocabulary models.

    Args:
        config_file (pathlib.Path):
            Location of model creator configuration file to specify input, output and MedCAT configuration.

    Returns:
        cdb (medcat.cdb.CDB):
            MedCAT concept database containing list of entities and synonyms, as well as context embeddings.
        vocab (medcat.vocab.Vocab):
            MedCAT vocabulary created from CDB and training documents.
    """
    # Load model creator configuration
    with open(config_file, 'r') as stream:
        config = yaml.safe_load(stream)

    # Load data for unsupervised training
    with open(Path(config['unsupervised_training_data_file']), 'r', encoding='utf-8') as training_data:
        training_data_list = [line.strip() for line in training_data]

    # Load MedCAT configuration
    medcat_config = Config()
    if 'medcat_config_file' in config:
        medcat_config.parse_config_file(Path(config['medcat_config_file']))

    # Create output dir if it does not exist
    output_dir = Path(config['output_dir'])
    output_dir.mkdir(parents=True, exist_ok=True)

    # Create models
    cdb = create_cdb(Path(config['concept_csv_file']), medcat_config)
    vocab = create_vocab(cdb, training_data_list, medcat_config, output_dir, config['unigram_table_size'])
    cdb = train_unsupervised(cdb, vocab, medcat_config, output_dir, training_data_list)

    return cdb, vocab


if __name__ == '__main__':
    # Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('config_file', help='YAML formatted file containing the parameters for model creator. An '
                                            'example can be found in `tests/model_creator/config_example.yml`')
    args = parser.parse_args()

    # Run pipeline
    create_models(args.config_file)



import setuptools

with open("./README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="medcat",
    setup_requires=["setuptools_scm"],
    use_scm_version={"local_scheme": "no-local-version", "fallback_version": "unknown"},
    author="w-is-h",
    author_email="w.kraljevic@gmail.com",
    description="Concept annotation tool for Electronic Health Records",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CogStack/MedCAT",
    packages=['medcat', 'medcat.utils', 'medcat.preprocessing', 'medcat.cogstack', 'medcat.ner', 'medcat.linking', 'medcat.datasets',
              'medcat.tokenizers', 'medcat.utils.meta_cat', 'medcat.pipeline', 'medcat.neo'],
    install_requires=[
        'numpy<1.22.9,>=1.21.4',
        'pandas<=1.3.4,>=1.1.5',
        'gensim~=4.1.2',
        'spacy<3.1.4,>=3.1.0',
        'scipy<=1.7.1,>=1.5.4',
        'transformers~=4.11.3',
        'torch<1.10,>=1.0',
        'tqdm>=4.27',
        'sklearn~=0.0',
        'elasticsearch>=7.10,<8.0.0',
        'dill~=0.3.4',
        'datasets~=1.14.0',
        'jsonpickle~=2.0.0',
        'psutil<6.0.0,>=5.8.0',
        'multiprocess', # seems to work better than standard mp
        'py2neo==2021.2.3',
        'aiofiles~=0.8.0',
        'ipywidgets~=7.6.5',
        'xxhash==2.0.2',
        'blis<=0.7.5',
        'click<=8.0.4', # Spacy breaks without this
        ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import types
import spacy
import gc
import logging
from typing import List, Optional, Union, Iterable, Callable
from multiprocessing import cpu_count
from spacy.tokens import Token, Doc, Span
from spacy.tokenizer import Tokenizer
from spacy.language import Language
from spacy.util import raise_error
from tqdm.autonotebook import tqdm
from medcat.linking.context_based_linker import Linker
from medcat.meta_cat import MetaCAT
from medcat.ner.vocab_based_ner import NER
from medcat.utils.normalizers import TokenNormalizer, BasicSpellChecker
from medcat.utils.loggers import add_handlers
from medcat.config import Config
from medcat.pipeline.pipe_runner import PipeRunner
from medcat.preprocessing.taggers import tag_skip_and_punct


class Pipe(object):
    r''' A wrapper around the standard spacy pipeline.

    Args:
        tokenizer (`spacy.tokenizer.Tokenizer`):
            What will be used to split text into tokens, can be anything built as a spacy tokenizer.
        config (`medcat.config.Config`):
            Global config for medcat.

    Properties:
        nlp (spacy.language.<lng>):
            The base spacy NLP pipeline.
    '''
    log = logging.getLogger(__package__)
    # Add file and console handlers
    log = add_handlers(log)

    def __init__(self, tokenizer: Tokenizer, config: Config) -> None:
        self._nlp = spacy.load(config.general['spacy_model'], disable=config.general['spacy_disabled_components'])
        if config.preprocessing['stopwords'] is not None:
            self._nlp.Defaults.stop_words = set(config.preprocessing['stopwords'])
        self._nlp.tokenizer = tokenizer(self._nlp, config)
        # Set max document length
        self._nlp.max_length = config.preprocessing.get('max_document_length', 1000000)
        self.config = config
        # Set log level
        self.log.setLevel(self.config.general['log_level'])

    def add_tagger(self, tagger: Callable, name: Optional[str] = None, additional_fields: List[str] = []) -> None:
        r''' Add any kind of a tagger for tokens.

        Args:
            tagger (`object/function`):
                Any object/function that takes a spacy doc as an input, does something
                and returns the same doc.
            name (`str`):
                Name for this component in the pipeline.
            additional_fields (`List[str]`):
                Fields to be added to the `_` properties of a token.
        '''
        component_factory_name = spacy.util.get_object_name(tagger)
        name = name if name is not None else component_factory_name
        Language.factory(name=component_factory_name, default_config={"config": self.config}, func=tagger)
        self._nlp.add_pipe(component_factory_name, name=name, first=True)

        # Add custom fields needed for this usecase
        Token.set_extension('to_skip', default=False, force=True)

        # Add any additional fields that are required
        for field in additional_fields:
            Token.set_extension(field, default=False, force=True)

    def add_token_normalizer(self, config: Config, name: Optional[str] = None, spell_checker: Optional[BasicSpellChecker] = None) -> None:
        token_normalizer = TokenNormalizer(config=config, spell_checker=spell_checker)
        component_name = spacy.util.get_object_name(token_normalizer)
        name = name if name is not None else component_name
        Language.component(name=component_name, func=token_normalizer)
        self._nlp.add_pipe(component_name, name=name, last=True)

        # Add custom fields needed for this usecase
        Token.set_extension('norm', default=None, force=True)

    def add_ner(self, ner: NER, name: Optional[str] = None) -> None:
        r''' Add NER from CAT to the pipeline, will also add the necessary fields
        to the document and Span objects.

        '''
        component_name = spacy.util.get_object_name(ner)
        name = name if name is not None else component_name
        Language.component(name=component_name, func=ner)
        self._nlp.add_pipe(component_name, name=name, last=True)

        Doc.set_extension('ents', default=[], force=True)
        Span.set_extension('confidence', default=-1, force=True)
        Span.set_extension('id', default=0, force=True)

        # Do not set this property if a vocabulary apporach is not used, this name must
        #refer to a name2cuis in the cdb.
        Span.set_extension('detected_name', default=None, force=True)
        Span.set_extension('link_candidates', default=None, force=True)

    def add_linker(self, linker: Linker, name: Optional[str] = None) -> None:
        r''' Add entity linker to the pipeline, will also add the necessary fields
        to Span object.

        linker (object/function):
            Any object/function created based on the requirements for a spaCy pipeline components. Have
            a look at https://spacy.io/usage/processing-pipelines#custom-components
        '''
        component_name = spacy.util.get_object_name(linker)
        name = name if name is not None else component_name
        Language.component(name=component_name, func=linker)
        self._nlp.add_pipe(component_name, name=name, last=True)
        Span.set_extension('cui', default=-1, force=True)
        Span.set_extension('context_similarity', default=-1, force=True)

    def add_meta_cat(self, meta_cat: MetaCAT, name: Optional[str] = None) -> None:
        component_name = spacy.util.get_object_name(meta_cat)
        name = name if name is not None else component_name
        Language.component(name=component_name, func=meta_cat)
        self._nlp.add_pipe(component_name, name=name, last=True)

        # meta_anns is a dictionary like {category_name: value, ...}
        Span.set_extension('meta_anns', default=None, force=True)
        # Used for sharing pre-processed data/tokens
        Doc.set_extension('share_tokens', default=None, force=True)

    def batch_multi_process(self,
                            texts: Iterable[str],
                            n_process: Optional[int] = None,
                            batch_size: Optional[int] = None) -> Iterable[Doc]:
        r''' Batch process a list of texts in parallel.

        Args:
            texts (`Iterable[str]`):
                The input sequence of texts to process.
            n_process (`int`):
                The number of processes running in parallel. Defaults to max(mp.cpu_count() - 1, 1).
            batch_size (`int`):
                The number of texts to buffer. Defaults to 1000.
            total (`int`):
                The number of texts in total.

        Return:
            Generator[Doc]:
                The output sequence of spacy documents with the extracted entities
        '''
        instance_name = "ensure_serializable"
        try:
            self._nlp.get_pipe(instance_name)
        except KeyError:
            component_name = spacy.util.get_object_name(self._ensure_serializable)
            Language.component(name=component_name, func=self._ensure_serializable)
            self._nlp.add_pipe(component_name, name=instance_name, last=True)

        n_process = n_process if n_process is not None else max(cpu_count() - 1, 1)
        batch_size = batch_size if batch_size is not None else 1000

        # If n_process < 0, multiprocessing will be either conducted inside pipeline components based on the con(when
        # 'parallel' is set to True) or not happen at all (when 'parallel' is set to False). Otherwise, multiprocessing
        # will be conducted at the pipeline level, i.e., texts will be processed sequentially by each pipeline component.
        if n_process < 0:
            inner_parallel = True
            n_process = 1
        else:
            inner_parallel = False

        component_cfg = {
            tag_skip_and_punct.name: {  # type: ignore
                'parallel': inner_parallel
            },
            TokenNormalizer.name: {
                'parallel': inner_parallel
            },
            NER.name: {
                'parallel': inner_parallel
            },
            Linker.name: {
                'parallel': inner_parallel
            }
        }

        return self._nlp.pipe(texts,    # type: ignore
                             n_process=n_process,
                             batch_size=batch_size,
                             component_cfg=component_cfg)

    def set_error_handler(self, error_handler: Callable) -> None:
        self._nlp.set_error_handler(error_handler)

    def reset_error_handler(self) -> None:
        self._nlp.set_error_handler(raise_error)

    def force_remove(self, component_name: str) -> None:
        try:
            self._nlp.remove_pipe(component_name)
        except ValueError:
            pass

    def destroy(self) -> None:
        del self._nlp
        gc.collect()

    @property
    def spacy_nlp(self) -> Language:
        """ The spaCy Language object
        """
        return self._nlp

    @staticmethod
    def _ensure_serializable(doc: Doc) -> Doc:
        return PipeRunner.serialize_entities(doc)

    def __call__(self, text: Union[str, Iterable[str]]) -> Union[Doc, List[Doc]]:
        if isinstance(text, str):
            return self._nlp(text) if len(text) > 0 else None
        elif isinstance(text, Iterable):
            docs = []
            for t in text if isinstance(text, types.GeneratorType) else tqdm(text, total=len(list(text))):
                try:
                    doc = self._nlp(t) if isinstance(t, str) and len(t) > 0 else None
                except Exception as e:
                    self.log.warning("Exception raised when processing text: %s", t[:50] + "..." if isinstance(t, str) else t)
                    self.log.warning(e, exc_info=True, stack_info=True)
                    doc = None
                docs.append(doc)
            return docs
        else:
            self.log.error("The input text should be either a string or a sequence of strings but got: %s", type(text))
            return None



# Generated by Django 3.2.10 on 2022-01-27 10:45

import django.core.files.storage
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('demo', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Downloader',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('first_name', models.CharField(max_length=20)),
                ('last_name', models.CharField(max_length=20)),
                ('email', models.EmailField(max_length=50)),
                ('affiliation', models.CharField(max_length=100)),
                ('funder', models.CharField(blank=True, default='', max_length=100)),
                ('use_case', models.TextField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='MedcatModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('model_name', models.CharField(max_length=20, unique=True)),
                ('model_file', models.FileField(storage=django.core.files.storage.FileSystemStorage(location='/medcat_data'), upload_to='')),
                ('model_display_name', models.CharField(max_length=50)),
            ],
        ),
    ]



import json
import pandas as pd


def original2concept_csv(data_path, out_path):
    f = open(data_path)
    csv_data = [['cui', 'type_id', 'name', 'name_status']]

    cui2names = {}

    for row in f:
        if row != '\n':
            if '|t|' not in row[0:13] and '|a|' not in row[0:13]:
                # Entity row
                parts = row.split("\t")
                cui = parts[5].strip()
                type_id = "|".join(parts[4].split(","))
                name = parts[3]
                csv_data.append([cui, type_id, name, 'A'])

                if cui in cui2names:
                    cui2names[cui].add(name)
                else:
                    cui2names[cui] = {name}

    df = pd.DataFrame(csv_data[1:], columns=csv_data[0])

    df.to_csv(out_path, index=False)

    return df


def original2pure_text(data_path, out_path):
    f = open(data_path)
    out = open(out_path, 'w')

    for row in f:
        if row != '\n':
            if '|t|' in row[0:13]:
                # It is title
                parts = row.split("|t|")
                title = parts[1].strip()
            elif '|a|' in row[0:13]:
                # Text row
                parts = row.split("|a|")
                text = parts[1].strip()
                out.write(title + " " + text + " " + "\n")
    out.close()


def original2json(data_path, out_path):
    f = open(data_path)
    data = {'projects': [{'name': 'medmentions', 'id': 0, 'documents': []}]}
    documents = []
    document = {}

    for row in f:
        if row != '\n':
            if '|t|' in row[0:13]:
                # It is title
                parts = row.split("|t|")
                title = parts[1].strip()
            elif '|a|' in row[0:13]:
                # Text row
                parts = row.split("|a|")
                text = parts[1].strip()
                document['text'] = title + " " + text
                document['annotations'] = []
            else:
                # Entity row
                parts = row.split("\t")
                start = int(parts[1])
                end = int(parts[2])
                cui = parts[5].strip()
                type_id = "|".join(parts[4].split(","))
                name = parts[3]

                document['annotations'].append({
                    'start': start,
                    'end': end,
                    'cui': cui,
                    'type_id': type_id,
                    'value': name})
        else:
            documents.append(document)
            document = {}
    data['projects'][0]['documents'] = documents

    json.dump(data, open(out_path, 'w'))
    return data



import logging
import os
import requests
import unittest
from spacy.lang.en import English
from medcat.preprocessing.tokenizers import spacy_split_all
from medcat.ner.vocab_based_ner import NER
from medcat.preprocessing.taggers import tag_skip_and_punct
from medcat.pipe import Pipe
from medcat.utils.normalizers import BasicSpellChecker
from medcat.vocab import Vocab
from medcat.preprocessing.cleaners import prepare_name
from medcat.linking.context_based_linker import Linker
from medcat.config import Config
from medcat.cdb import CDB


class A_NERTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("Set up CDB")
        cls.config = Config()
        cls.config.general['log_level'] = logging.INFO
        cls.config.general["spacy_model"] = "en_core_web_md"
        cls.cdb = CDB(config=cls.config)

        print("Set up Vocab")
        vocab_path = "./tmp_vocab.dat"
        if not os.path.exists(vocab_path):
            tmp = requests.get("https://medcat.rosalind.kcl.ac.uk/media/vocab.dat")
            with open(vocab_path, 'wb') as f:
                f.write(tmp.content)

        cls.vocab = Vocab.load(vocab_path)


        print("Set up NLP pipeline")
        cls.pipe = Pipe(tokenizer=spacy_split_all, config=cls.config)
        cls.pipe.add_tagger(tagger=tag_skip_and_punct,
                            name='skip_and_punct',
                            additional_fields=['is_punct'])

        cls.spell_checker = BasicSpellChecker(cdb_vocab=cls.cdb.vocab, config=cls.config, data_vocab=cls.vocab)
        cls.pipe.add_token_normalizer(spell_checker=cls.spell_checker, config=cls.config)
        cls.ner = NER(cls.cdb, cls.config)
        cls.pipe.add_ner(cls.ner)

        print("Set up Linker")
        cls.link = Linker(cls.cdb, cls.vocab, cls.config)
        cls.pipe.add_linker(cls.link)

        print("Set limits for tokens and uppercase")
        cls.config.ner['max_skip_tokens'] = 1
        cls.config.ner['upper_case_limit_len'] = 4
        cls.config.linking['disamb_length_limit'] = 2
        cls.config.general["spacy_model"] = "en_core_sci_sm"

        print("Add concepts")
        cls.cdb.add_names(cui='S-229004', names=prepare_name('Movar', cls.pipe, {}, cls.config))
        cls.cdb.add_names(cui='S-229004', names=prepare_name('Movar viruses', cls.pipe, {}, cls.config))
        cls.cdb.add_names(cui='S-229005', names=prepare_name('CDB', cls.pipe, {}, cls.config))

        print("Add test text")
        cls.text = "CDB - I was running and then Movar    Viruses attacked and CDb"
        cls.text_post_pipe = cls.pipe(cls.text)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.pipe.destroy()

    def test_aa_cdb_names_output(self):
        target_result = {'S-229004': {'movar~viruse', 'movar', 'movar~viruses'}, 'S-229005': {'cdb'}}
        self.assertEqual(self.cdb.cui2names, target_result)

    def test_ab_entities_length(self):
        self.assertEqual(len(self.text_post_pipe._.ents), 2, "Should equal 2")

    def test_ac_entities_linked_candidates(self):
        target_result = 'S-229004'
        self.assertEqual(self.text_post_pipe._.ents[0]._.link_candidates[0], target_result)

    def test_ad_max_skip_entities_length(self):
        self.config.ner['max_skip_tokens'] = 3
        self.text_post_pipe = self.pipe(self.text)
        self.assertEqual(len(self.text_post_pipe._.ents), 3, "Should equal 3")

    def test_ae_upper_case_entities_length(self):
        self.config.ner['upper_case_limit_len'] = 3
        self.text_post_pipe = self.pipe(self.text)
        self.assertEqual(len(self.text_post_pipe._.ents), 4, "Should equal 4")

    def test_af_min_name_entities_length(self):
        self.config.ner['min_name_len'] = 4
        self.text_post_pipe = self.pipe(self.text)
        print(self.text)
        print(self.text_post_pipe._.ents)
        self.assertEqual(len(self.text_post_pipe._.ents), 2, "Should equal 2")


if __name__ == '__main__':
    unittest.main()
